# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import pathlib
import sys


print("python exec:", sys.executable)
print("sys.path:", sys.path)
root = pathlib.Path(__file__).parent.parent.absolute()
os.environ["PYTHONPATH"] = str(root)
sys.path.insert(0, str(root))

import xmip  # isort:skip
from importlib.metadata import version  # isort:skip

release = version("xmip")
# for example take major/minor/patch
version = ".".join(release.split(".")[:3])

# From https://github.com/pypa/setuptools_scm/#usage-from-sphinx

# -- Project information -----------------------------------------------------

project = "xmip"
copyright = "2021, xmip maintainers"
author = "xmip maintainers"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "nbsphinx",
    "recommonmark",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "numpydoc",
    "nbsphinx",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    "sphinxcontrib.srclinks",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "**.ipynb_checkpoints", "Thumbs.db", ".DS_Store"]

# link to github issues
extlinks = {
    "issue": ("https://github.com/jbusecke/xmip/issues/%s", "GH#"),
    "pull": ("https://github.com/jbusecke/xmip/issues/%s", "GH#"),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pangeo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



import warnings

import numpy as np
import pkg_resources
import xarray as xr
import yaml

from xgcm import Grid
from xgcm.autogenerate import generate_grid_ds


path = "specs/staggered_grid_config.yaml"  # always use slash
grid_spec = pkg_resources.resource_filename(__name__, path)


def _parse_bounds_vertex(da, dim="bnds", position=[0, 1]):
    """Convenience function to extract positions from bounds/verticies"""
    return tuple([da.isel({dim: i}).load().data for i in position])


def _interp_vertex_to_bounds(da, orientation):
    """
    Convenience function to average 4 vertex points into two bound points.
    Helpful to recreate e.g. the latitude at the `lon_bounds` points.
    """
    if orientation == "x":
        datasets = [
            da.isel(vertex=[0, 1]).mean("vertex"),
            da.isel(vertex=[3, 2]).mean("vertex"),
        ]
    elif orientation == "y":
        datasets = [
            da.isel(vertex=[0, 3]).mean("vertex"),
            da.isel(vertex=[1, 2]).mean("vertex"),
        ]

    return xr.concat(datasets, dim="bnds")


def distance_deg(lon0, lat0, lon1, lat1):
    """Calculate the distance in degress longitude and latitude between two points

    Parameters
    ----------
    lon0 : np.array
        Longitude of first point
    lat0 : np.array
        Latitude of first point
    lon1 : np.array
        Longitude of second point
    lat1 : np.array
        Latitude of second point
    """
    delta_lon = lon1 - lon0
    delta_lat = lat1 - lat0
    # very small differences can end up negative, so zero them out based on a simple
    # criterion
    # this should work for CMIP6 (no 1/1 deg models) but should be based on actual grid
    # info in the future
    small_crit = 1 / 10
    delta_lon = np.where(
        abs(delta_lon) < small_crit, 0.0, delta_lon
    )  # , np.nan, delta_lon)
    delta_lat = np.where(
        abs(delta_lat) < small_crit, 0.0, delta_lat
    )  # , np.nan, delta_lat)

    #     # some bounds are wrapped aroud the lon discontinuty.
    delta_lon = np.where(delta_lon < (-small_crit * 2), 360 + delta_lon, delta_lon)  #
    delta_lon = np.where(
        delta_lon > (360 + small_crit * 2), -360 + delta_lon, delta_lon
    )

    return delta_lon, delta_lat


def distance(lon0, lat0, lon1, lat1):
    """Calculate the distance in m between two points on a spherical globe

    Parameters
    ----------
    lon0 : np.array
        Longitude of first point
    lat0 : np.array
        Latitude of first point
    lon1 : np.array
        Longitude of second point
    lat1 : np.array
        Latitude of second point
    """
    Re = 6.378e6
    delta_lon, delta_lat = distance_deg(lon0, lat0, lon1, lat1)
    dy = Re * (np.pi * delta_lat / 180)
    dx = Re * (np.pi * delta_lon / 180) * np.cos(np.pi * lat0 / 180)
    return np.sqrt(dx**2 + dy**2)


def recreate_metrics(ds, grid):
    """Recreate a full set of horizontal distance metrics.

    Calculates distances between points in lon/lat coordinates


    The naming of the metrics is as follows:
    [metric_axis]_t : metric centered at tracer point
    [metric_axis]_gx : metric at the cell face on the x-axis.
        For instance `dx_gx` is the x distance centered on the eastern cell face if the shift is `right`
    [metric_axis]_gy : As above but along the y-axis
    [metric_axis]_gxgy : The metric located at the corner point.
        For example `dy_dxdy` is the y distance on the south-west corner if both axes as shifted left.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset.
    grid : xgcm.Grid
        xgcm Grid object matching `ds`

    Returns
    -------
    xr.Dataset, dict
        Dataset with added metrics as coordinates and dictionary that can be passed to xgcm.Grid to recognize new metrics
    """
    ds = ds.copy()

    # Since this puts out numpy arrays, the arrays need to be transposed correctly
    transpose_dims = ["y", "x"]
    dims = [di for di in ds.dims if di not in transpose_dims]

    ds = ds.transpose(*tuple(transpose_dims + dims))

    # is the vel point on left or right?
    axis_vel_pos = {
        axis: list(set(grid.axes[axis].coords.keys()) - set(["center"]))[0]
        for axis in ["X", "Y"]
    }
    # determine the appropriate vertex position for the north/south and east/west edge,
    # based on the grid config
    if axis_vel_pos["Y"] in ["left"]:
        ns_vertex_idx = [0, 3]
        ns_bound_idx = [0]
    elif axis_vel_pos["Y"] in ["right"]:
        ns_vertex_idx = [1, 2]
        ns_bound_idx = [1]

    if axis_vel_pos["X"] in ["left"]:
        ew_vertex_idx = [0, 1]
        ew_bound_idx = [0]
    elif axis_vel_pos["X"] in ["right"]:
        ew_vertex_idx = [3, 2]
        ew_bound_idx = [1]

    # infer dx at tracer points
    if "lon_bounds" in ds.coords and "lat_verticies" in ds.coords:
        lon0, lon1 = _parse_bounds_vertex(ds["lon_bounds"])
        lat0, lat1 = _parse_bounds_vertex(
            _interp_vertex_to_bounds(ds["lat_verticies"], "x")
        )
        dist = distance(lon0, lat0, lon1, lat1)
        ds.coords["dx_t"] = xr.DataArray(dist, coords=ds.lon.coords)

    # infer dy at tracer points
    if "lat_bounds" in ds.coords and "lon_verticies" in ds.coords:
        lat0, lat1 = _parse_bounds_vertex(ds["lat_bounds"])
        lon0, lon1 = _parse_bounds_vertex(
            _interp_vertex_to_bounds(ds["lon_verticies"], "y")
        )
        dist = distance(lon0, lat0, lon1, lat1)
        ds.coords["dy_t"] = xr.DataArray(dist, coords=ds.lon.coords)

    if "lon_verticies" in ds.coords and "lat_verticies" in ds.coords:

        # infer dx at the north/south face
        lon0, lon1 = _parse_bounds_vertex(
            ds["lon_verticies"], dim="vertex", position=ns_vertex_idx
        )
        lat0, lat1 = _parse_bounds_vertex(
            ds["lat_verticies"], dim="vertex", position=ns_vertex_idx
        )
        dist = distance(lon0, lat0, lon1, lat1)
        ds.coords["dx_gy"] = xr.DataArray(
            dist, coords=grid.interp(ds.lon, "Y", boundary="extrapolate").coords
        )

        # infer dy at the east/west face
        lon0, lon1 = _parse_bounds_vertex(
            ds["lon_verticies"], dim="vertex", position=ew_vertex_idx
        )
        lat0, lat1 = _parse_bounds_vertex(
            ds["lat_verticies"], dim="vertex", position=ew_vertex_idx
        )
        dist = distance(lon0, lat0, lon1, lat1)
        ds.coords["dy_gx"] = xr.DataArray(
            dist, coords=grid.interp(ds.lon, "X", boundary="extrapolate").coords
        )

    # for the distances that dont line up with the cell boundaries we need some different logic
    boundary = "extend"
    # TODO: This should be removed once we have the default boundary merged in xgcm

    # infer dx at eastern/western bound from tracer points
    lon0, lon1 = grid.axes["X"]._get_neighbor_data_pairs(
        ds.lon.load(), axis_vel_pos["X"]
    )
    lat0, lat1 = grid.axes["X"]._get_neighbor_data_pairs(
        ds.lat.load(), axis_vel_pos["X"]
    )
    dx = distance(lon0, lat0, lon1, lat1)
    ds.coords["dx_gx"] = xr.DataArray(
        dx, coords=grid.interp(ds.lon, "X", boundary=boundary).coords
    )

    # infer dy at northern bound from tracer points
    lat0, lat1 = grid.axes["Y"]._get_neighbor_data_pairs(
        ds.lat.load(), axis_vel_pos["Y"], boundary=boundary
    )
    lon0, lon1 = grid.axes["Y"]._get_neighbor_data_pairs(
        ds.lon.load(), axis_vel_pos["Y"], boundary=boundary
    )
    dy = distance(lon0, lat0, lon1, lat1)
    ds.coords["dy_gy"] = xr.DataArray(
        dy, coords=grid.interp(ds.lat, "Y", boundary=boundary).coords
    )

    # infer dx at the corner point
    lon0, lon1 = grid.axes["X"]._get_neighbor_data_pairs(
        _interp_vertex_to_bounds(ds.lon_verticies.load(), "y")
        .isel(bnds=ns_bound_idx)
        .squeeze(),
        axis_vel_pos["X"],
    )
    lat0, lat1 = grid.axes["X"]._get_neighbor_data_pairs(
        ds.lat_bounds.isel(bnds=ns_bound_idx).squeeze().load(), axis_vel_pos["X"]
    )
    dx = distance(lon0, lat0, lon1, lat1)
    ds.coords["dx_gxgy"] = xr.DataArray(
        dx,
        coords=grid.interp(
            grid.interp(ds.lon, "X", boundary=boundary), "Y", boundary=boundary
        ).coords,
    )

    # infer dy at the corner point
    lat0, lat1 = grid.axes["Y"]._get_neighbor_data_pairs(
        _interp_vertex_to_bounds(ds.lat_verticies.load(), "x")
        .isel(bnds=ew_bound_idx)
        .squeeze(),
        axis_vel_pos["Y"],
    )
    lon0, lon1 = grid.axes["Y"]._get_neighbor_data_pairs(
        ds.lon_bounds.isel(bnds=ew_bound_idx).squeeze().load(), axis_vel_pos["Y"]
    )
    dy = distance(lon0, lat0, lon1, lat1)
    ds.coords["dy_gxgy"] = xr.DataArray(
        dy,
        coords=grid.interp(
            grid.interp(ds.lon, "X", boundary=boundary), "Y", boundary=boundary
        ).coords,
    )

    # infer dz at tracer point
    if "lev_bounds" in ds.coords:
        ds = ds.assign_coords(
            dz_t=("lev", ds["lev_bounds"].diff("bnds").squeeze(drop=True).data)
        )

    metrics_dict = {
        "X": [co for co in ["dx_t", "dx_gy", "dx_gx"] if co in ds.coords],
        "Y": [co for co in ["dy_t", "dy_gy", "dy_gx"] if co in ds.coords],
        "Z": [co for co in ["dz_t"] if co in ds.coords],
    }
    # # only put out axes that have entries
    metrics_dict = {k: v for k, v in metrics_dict.items() if len(v) > 0}

    return ds, metrics_dict


def detect_shift(ds_base, ds, axis):
    """Detects the shift of `ds` relative to `ds` on logical grid axes, using
    lon and lat positions.

    Parameters
    ----------
    ds_base : xr.Dataset
        Reference ('base') dataset to compare to. Assumed that this is located at the 'center' coordinate.
    ds : xr.Dataset
        Comparison dataset. The resulting shift will be computed as this dataset relative to `ds_base`
    axis : str
        xgcm logical axis on which to detect the shift

    Returns
    -------
    str
        Shift string output, in xgcm conventions.
    """
    ds_base = ds_base.copy()
    ds = ds.copy()
    axis = axis.lower()
    axis_coords = {"x": "lon", "y": "lat"}

    # check the shift only for one point, somewhat in the center to avoid the
    # distorted polar regions
    check_point = {"x": len(ds_base.x) // 2, "y": len(ds_base.y) // 2}
    check_point_diff = {k: [v, v + 1] for k, v in check_point.items()}

    shift = (
        ds.isel(**check_point)[axis_coords[axis]].load().data
        - ds_base.isel(**check_point)[axis_coords[axis]].load().data
    )
    diff = ds[axis].isel({axis: check_point_diff[axis]}).diff(axis).data.tolist()[0]
    threshold = 0.1
    # the fraction of full cell distance, that a point has to be shifted in order to
    # be recognized.
    # This avoids detection of shifts for very small differences that sometimes happen
    # if the coordinates were written e.g. by different modulel of a model

    axis_shift = "center"

    if shift > (diff * threshold):
        axis_shift = "right"
    elif shift < -(diff * threshold):
        axis_shift = "left"
    return axis_shift


def create_full_grid(base_ds, grid_dict=None):
    """Generate a full xgcm-compatible dataset from a reference datasets `base_ds`.
    This dataset should be representing a tracer fields, e.g. the cell center.

    Parameters
    ----------
    base_ds : xr.Dataset
        The reference ('base') datasets, assumed to be at the tracer position/cell center
    grid_dict : dict, optional
        Dictionary with info about the grid staggering.
        Must be encoded using the base_ds attrs (e.g. {'model_name':{'axis_shift':{'X':'left',...}}}).
        If deactivated (default), will load from the internal database for CMIP6 models, by default None

    Returns
    -------
    xr.Dataset
        xgcm compatible dataset
    """

    # load dict with grid shift info for each axis
    if grid_dict is None:
        ff = open(grid_spec, "r")
        grid_dict = yaml.safe_load(ff)
        ff.close()

    source_id = base_ds.attrs["source_id"]
    grid_label = base_ds.attrs["grid_label"]

    # if source_id not in dict, and grid label is gn, warn and ask to submit an issue
    try:
        axis_shift = grid_dict[source_id][grid_label]["axis_shift"]
    except KeyError:
        warnings.warn(
            f"Could not find the source_id/grid_label ({source_id}/{grid_label}) combo in `grid_dict`, returning `None`. Please submit an issue to github: https://github.com/jbusecke/xmip/issues"
        )
        return None

    position = {k: ("center", axis_shift[k]) for k in axis_shift.keys()}

    axis_dict = {"X": "x", "Y": "y"}

    ds_grid = generate_grid_ds(
        base_ds, axis_dict, position=position, boundary_discontinuity={"X": 360}
    )

    # TODO: man parse lev and lev_bounds as center and outer dims.
    # I should also be able to do this with `generate_grid_ds`, but here we
    # have the `lev_bounds` with most models, so that is probably more reliable.
    # cheapest solution right now
    if "lev" in ds_grid.dims:
        ds_grid["lev"].attrs["axis"] = "Z"

    return ds_grid


def combine_staggered_grid(
    ds_base, other_ds=None, recalculate_metrics=False, grid_dict=None, **kwargs
):
    """Combine a reference datasets with a list of other datasets to a full xgcm-compatible staggered grid datasets.


    Parameters
    ----------
    ds_base : xr.Dataset
        The reference ('base') datasets, assumed to be at the tracer position/cell center
    other_ds : list,xr.Dataset, optional
        List of datasets representing different variables. Their grid position will be
        automatically detected relative to `ds_base`. Coordinates and attrs of these added datasets will be lost
        , by default None
    recalculate_metrics : bool, optional
        nables the reconstruction of grid metrics usign simple
        spherical geometry, by default False

        !!! Check your results carefully when using reconstructed values,
        these might differe substantially if the grid geometry is complicated.
    grid_dict : dict, optional
        Dictionary for staggered grid setup. See `create_full_grid` for detauls
        If None (default), will load staggered grid info from internal database, by default None

    Returns
    -------
    xr.Dataset
        Single xgcm-compatible dataset, containing all variables on their respective staggered grid position.
    """
    ds_base = ds_base.copy()
    if isinstance(other_ds, xr.Dataset):
        other_ds = [other_ds]

    ds_g = create_full_grid(ds_base, grid_dict=grid_dict)

    if ds_g is None:
        warnings.warn("Staggered Grid creation failed. Returning `None`")
        return None, None

    # save attrs out for later (something during alignment destroys them)
    dim_attrs_dict = {}
    for di in ds_g.dims:
        dim_attrs_dict[di] = ds_g[di].attrs

    # TODO: metrics and interpolation of metrics if they are parsed

    # parse other variables
    if other_ds is not None:
        for ds_new in other_ds:
            ds_new = ds_new.copy()
            # strip everything but the variable_id (perhaps I would want to
            # loosen this in the future)
            ds_new = ds_new[ds_new.attrs["variable_id"]]

            if not all(
                [
                    len(ds_new[di]) == len(ds_g[di])
                    for di in ds_new.dims
                    if di not in ["member_id", "time"]
                ]
            ):
                warnings.warn(
                    f"Could not parse `{ds_new.name}`, due to a size mismatch. If this is the MRI model, the grid convention is currently not supported."
                )
            else:
                # detect shift and rename accordingly
                rename_dict = {}
                for axis in ["X", "Y"]:
                    shift = detect_shift(ds_base, ds_new, axis)

                    if shift != "center":
                        rename_dict[axis.lower()] = axis.lower() + "_" + shift
                ds_new = ds_new.rename(rename_dict)
                ds_new = ds_new.reset_coords(drop=True)
                # TODO: This needs to be coded more generally, for now hardcode x and y
                force_align_dims = [di for di in ds_new.dims if "x" in di or "y" in di]
                _, ds_new = xr.align(
                    ds_g.copy(),
                    ds_new,
                    join="override",
                    exclude=[di for di in ds_new.dims if di not in force_align_dims],
                )
                additional_dims = [di for di in ds_new.dims if di not in ds_g.dims]
                if len(additional_dims) > 0:
                    raise RuntimeError(
                        f"While trying to parse `{ds_new.name}`, detected dims that are not in the base dataset:[{additional_dims}]"
                    )
                ds_g[ds_new.name] = ds_new

    # Restore dims attrs from the beginning
    for di in ds_g.dims:
        ds_g.coords[di].attrs.update(dim_attrs_dict[di])

    grid_kwargs = {"periodic": ["X"]}
    grid_kwargs.update(kwargs)
    grid = Grid(ds_g, grid_kwargs)

    # if activated calculate metrics
    if recalculate_metrics:
        grid_kwargs.pop(
            "metrics", None
        )  # remove any passed metrics when recalculating them
        # I might be able to refine this more to e.g. allow axes that are not recreated.

        ds_g, metrics_dict = recreate_metrics(ds_g, grid)
        # this might fail in circumstances, where the
        grid_kwargs["metrics"] = metrics_dict
        grid = Grid(ds_g, **grid_kwargs)
    return grid, ds_g



import pytest
import xarray as xr

from xmip.utils import cmip6_dataset_id, google_cmip_col, model_id_match


def test_google_cmip_col():
    try:
        import intake
    except ImportError:
        intake = None
    if intake is None:
        with pytest.raises(ImportError):
            col = google_cmip_col(catalog="main")
    else:
        col = google_cmip_col(catalog="main")
        assert (
            col.catalog_file == "https://storage.googleapis.com/cmip6/pangeo-cmip6.csv"
        )

        with pytest.raises(ValueError):
            col = google_cmip_col(catalog="wrong")


def test_model_id_match():

    # wrong amount of elements
    with pytest.raises(ValueError):

        model_id_match([("A", "a", "aa"), ("A", "a", "aa", "aaa")], ("A", "a", "aa"))

    with pytest.raises(ValueError):
        model_id_match([("A", "a", "aa"), ("B", "a", "aa")], ("A", "a", "aa", "aaa"))

    assert model_id_match([("A", "a", "aa"), ("B", "a", "aa")], ("A", "a", "aa"))
    assert ~model_id_match([("A", ["b", "c"], "aa")], ("A", "a", "aa"))
    assert ~model_id_match([("A", ["b", "c"], "aa")], ("A", "a", "aa"))
    assert ~model_id_match(
        [("EC-Earth3-AerChem", ["so"], "historical", "gn")],
        ("EC-Earth3", ["so"], "historical", "gn"),
    )
    assert ~model_id_match([("A", "a", "aa"), ("B", "a", "aa")], ("AA", "a", "aa"))
    assert ~model_id_match([("AA", "a", "aa"), ("B", "a", "aa")], ("A", "a", "aa"))
    assert ~model_id_match([(["AA"], "a", "aa"), ("B", "a", "aa")], ("A", "a", "aa"))
    assert ~model_id_match([(["AA"], ["a"], "aa"), ("B", "a", "aa")], ("A", "a", "aa"))
    assert model_id_match([("*", "a", "aa")], ("whatever", "a", "aa"))
    assert model_id_match([(["bb", "b"], "a", "aa")], ("b", "a", "aa"))
    assert model_id_match(
        [(["bb", "b"], "a", "aa"), (["bb", "b"], "c", "cc")], ("bb", "a", "aa")
    )


def test_cmip6_dataset_id():
    ds = xr.Dataset({"data": 4})

    ds.attrs = {
        "activity_id": "ai",
        "institution_id": "ii",
        "source_id": "si",
        "variant_label": "vl",
        "experiment_id": "ei",
        "table_id": "ti",
        "grid_label": "gl",
        "variable_id": "vi",
    }

    assert cmip6_dataset_id(ds) == "ai.ii.si.ei.vl.ti.gl.none.vi"
    assert cmip6_dataset_id(ds, sep="_") == "ai_ii_si_ei_vl_ti_gl_none_vi"
    assert (
        cmip6_dataset_id(ds, id_attrs=["grid_label", "activity_id", "wrong_attrs"])
        == "gl.ai.none"
    )



import itertools

import numpy as np
import pytest
import xarray as xr

from xmip.postprocessing import EXACT_ATTRS
from xmip.preprocessing import (
    broadcast_lonlat,
    cmip6_renaming_dict,
    combined_preprocessing,
    correct_coordinates,
    correct_lon,
    correct_units,
    fix_metadata,
    maybe_convert_bounds_to_vertex,
    maybe_convert_vertex_to_bounds,
    parse_lon_lat_bounds,
    promote_empty_dims,
    rename_cmip6,
    replace_x_y_nominal_lat_lon,
    sort_vertex_order,
)


def create_test_ds(xname, yname, zname, xlen, ylen, zlen):
    x = np.linspace(0, 359, xlen)
    y = np.linspace(-90, 89, ylen)
    z = np.linspace(0, 5000, zlen)

    data = np.random.rand(len(x), len(y), len(z))
    ds = xr.DataArray(data, coords=[(xname, x), (yname, y), (zname, z)]).to_dataset(
        name="test"
    )
    ds.attrs["source_id"] = "test_id"
    # if x and y are not lon and lat, add lon and lat to make sure there are no conflicts
    lon = ds[xname] * xr.ones_like(ds[yname])
    lat = xr.ones_like(ds[xname]) * ds[yname]
    if xname != "lon" and yname != "lat":
        ds = ds.assign_coords(lon=lon, lat=lat)
    else:
        ds = ds.assign_coords(longitude=lon, latitude=lat)
    return ds


@pytest.mark.parametrize("xname", ["i", "x"])
@pytest.mark.parametrize("yname", ["j", "y"])
@pytest.mark.parametrize("zname", ["lev", "olev", "olevel"])
@pytest.mark.parametrize("missing_dim", [None, "x", "y", "z"])
def test_rename_cmip6(xname, yname, zname, missing_dim):
    xlen, ylen, zlen = (10, 5, 6)
    ds = create_test_ds(xname, yname, zname, xlen, ylen, zlen)

    if missing_dim == "x":
        ds = ds.isel({xname: 0}).squeeze()
    elif missing_dim == "y":
        ds = ds.isel({yname: 0}).squeeze()
    elif missing_dim == "z":
        ds = ds.isel({zname: 0}).squeeze()

    ds_renamed = rename_cmip6(ds, cmip6_renaming_dict())
    assert set(ds_renamed.dims).issubset(set(["x", "y", "lev"]))
    if not missing_dim == "x":
        assert xlen == len(ds_renamed.x)
    if not missing_dim == "y":
        assert ylen == len(ds_renamed.y)
    if not missing_dim == "z":
        assert zlen == len(ds_renamed.lev)


@pytest.mark.parametrize("xname", ["i", "x"])
@pytest.mark.parametrize("yname", ["j", "y"])
def test_rename_cmip6_worst_case(xname, yname):
    xlen, ylen, zlen = (10, 5, 6)
    ds = create_test_ds(xname, yname, "lev", xlen, ylen, zlen)

    print(ds.lon)
    # now rename only some of the coordinates to the correct naming
    ds = ds.assign_coords(
        {"lon": ds.lon.reset_coords(drop=True).rename({xname: "x", yname: "y"})}
    )
    ds_renamed = rename_cmip6(ds, cmip6_renaming_dict())

    assert set(ds_renamed.dims) == set(["x", "y", "lev"])


def test_broadcast_lonlat():
    x = np.arange(-180, 179, 5)
    y = np.arange(-90, 90, 6)
    data = np.random.rand(len(x), len(y))
    ds = xr.DataArray(data, dims=["x", "y"], coords={"x": x, "y": y}).to_dataset(
        name="test"
    )
    expected = ds.copy()
    expected.coords["lon"] = ds.x * xr.ones_like(ds.y)
    expected.coords["lat"] = xr.ones_like(ds.x) * ds.y

    ds_test = broadcast_lonlat(ds)
    xr.testing.assert_identical(expected, ds_test)


def test_promote_empty_dims():
    xlen, ylen, zlen = (10, 5, 6)
    ds = create_test_ds("x", "y", "z", xlen, ylen, zlen)
    ds = ds.drop_vars(["x", "y", "z"])
    ds_promoted = promote_empty_dims(ds)
    assert set(["x", "y", "z"]).issubset(set(ds_promoted.coords))


@pytest.mark.parametrize("nans", [True, False])
@pytest.mark.parametrize("dask", [True, False])
def test_replace_x_y_nominal_lat_lon(dask, nans):
    x = np.linspace(0, 720, 10)
    y = np.linspace(-200, 140, 5)
    lon = xr.DataArray(np.linspace(0, 360, len(x)), coords=[("x", x)])
    lat = xr.DataArray(np.linspace(-90, 90, len(y)), coords=[("y", y)])
    llon = lon * xr.ones_like(lat)
    llat = xr.ones_like(lon) * lat

    data = np.random.rand(len(x), len(y))
    ds = xr.DataArray(data, coords=[("x", x), ("y", y)]).to_dataset(name="data")
    ds.coords["lon"] = llon
    ds.coords["lat"] = llat

    if nans:
        lon = ds["lon"].load().data
        lon[0, :] = np.nan
        lon[-1, :] = np.nan
        lon[:, 0] = np.nan
        lon[:, -1] = np.nan
        lon[15:23, 23:26] = np.nan
        ds["lon"].data = lon

        # for lats put only some nans in the middle.
        # I currently have no way to interpolate lats at the edge.
        lat = ds["lat"].load().data
        lat[15:23, 23:26] = np.nan
        ds["lat"].data = lat

    if dask:
        ds = ds.chunk({"x": -1, "y": -1})
        ds.coords["lon"] = ds.coords["lon"].chunk({"x": -1, "y": -1})
        ds.coords["lat"] = ds.coords["lat"].chunk({"x": -1, "y": -1})

    replaced_ds = replace_x_y_nominal_lat_lon(ds)

    assert all(~np.isnan(replaced_ds.x))
    assert all(~np.isnan(replaced_ds.y))

    assert all(replaced_ds.x.diff("x") > 0)
    assert all(replaced_ds.y.diff("y") > 0)
    assert len(replaced_ds.lon.shape) == 2
    assert len(replaced_ds.lat.shape) == 2
    assert set(replaced_ds.lon.dims) == set(["x", "y"])
    assert set(replaced_ds.lat.dims) == set(["x", "y"])
    assert all(~np.isnan(replaced_ds.x))
    assert all(~np.isnan(replaced_ds.y))

    # test a dataset that would result in duplicates with current method
    x = np.linspace(0, 720, 4)
    y = np.linspace(-200, 140, 3)
    llon = xr.DataArray(
        np.array([[0, 50, 100, 150], [0, 50, 100, 150], [0, 50, 100, 150]]),
        coords=[("y", y), ("x", x)],
    )
    llat = xr.DataArray(
        np.array([[0, 0, 10, 0], [10, 0, 0, 0], [20, 20, 20, 20]]),
        coords=[("y", y), ("x", x)],
    )
    data = np.random.rand(len(x), len(y))
    ds = xr.DataArray(data, coords=[("x", x), ("y", y)]).to_dataset(name="data")
    ds.coords["lon"] = llon
    ds.coords["lat"] = llat

    if dask:
        ds = ds.chunk({"x": -1, "y": -1})
        ds.coords["lon"] = ds.coords["lon"].chunk({"x": -1, "y": -1})
        ds.coords["lat"] = ds.coords["lat"].chunk({"x": -1, "y": -1})

    replaced_ds = replace_x_y_nominal_lat_lon(ds)
    assert all(~np.isnan(replaced_ds.x))
    assert all(~np.isnan(replaced_ds.y))
    assert len(replaced_ds.y) == len(np.unique(replaced_ds.y))
    assert len(replaced_ds.x) == len(np.unique(replaced_ds.x))
    # make sure values are sorted in ascending order
    assert all(replaced_ds.x.diff("x") > 0)
    assert all(replaced_ds.y.diff("y") > 0)
    assert len(replaced_ds.lon.shape) == 2
    assert len(replaced_ds.lat.shape) == 2
    assert set(replaced_ds.lon.dims) == set(["x", "y"])
    assert set(replaced_ds.lat.dims) == set(["x", "y"])


@pytest.mark.parametrize(
    "coord",
    [
        "x",
        "y",
        "lon",
        "lat",
        "lev",
        "lev_bounds",
        "lon_bounds",
        "lat_bounds",
        "time_bounds",
        "lat_verticies",
        "lon_verticies",
    ],
)
def test_correct_coordinates(coord):
    xlen, ylen, zlen = (10, 5, 6)
    ds = create_test_ds("xx", "yy", "zz", xlen, ylen, zlen)
    # set a new variable
    ds = ds.assign({coord: ds.test})

    ds_corrected = correct_coordinates(ds)
    assert coord in list(ds_corrected.coords)


def test_parse_lon_lat_bounds():
    lon = np.arange(0, 10)
    lat = np.arange(20, 30)
    data = np.random.rand(len(lon), len(lat))
    ds = xr.DataArray(data, dims=["x", "y"], coords={"x": lon, "y": lat}).to_dataset(
        name="test"
    )
    ds.coords["lon"] = ds.x * xr.ones_like(ds.y)
    ds.coords["lat"] = xr.ones_like(ds.x) * ds.y

    ds.coords["lon_bounds"] = (
        xr.DataArray([-0.1, -0.1, 0.1, 0.1], dims=["vertex"]) + ds["lon"]
    )
    ds.coords["lat_bounds"] = (
        xr.DataArray([-0.1, 0.1, 0.1, -0.1], dims=["vertex"]) + ds["lat"]
    )

    ds_test = parse_lon_lat_bounds(ds)
    assert "lon_verticies" in ds_test.coords
    assert "lat_verticies" in ds_test.coords

    # introduce a time diemension
    for wrong_coord in ["lon_bounds", "lat_bounds"]:
        # TODO: this should also test lev_bounds.
        # Are there other coords that should be purged of the
        ds_wrong = ds.copy()
        ds_wrong.coords[wrong_coord] = ds_wrong.coords[wrong_coord] * xr.DataArray(
            range(5), dims=["time"]
        )

        ds_test2 = parse_lon_lat_bounds(ds_wrong)
        assert "time" in ds_wrong.dims
        assert "time" not in ds_test2.variables


@pytest.mark.parametrize("missing_values", [False, 1e36, -1e36, 1001, -1001])
@pytest.mark.parametrize(
    "shift",
    [
        -70,
        -180,
        -360,
    ],
)  # cant handle positive shifts yet
def test_correct_lon(missing_values, shift):
    xlen, ylen, zlen = (40, 20, 6)
    ds = create_test_ds("x", "y", "lev", xlen, ylen, zlen)
    ds = ds.assign_coords(x=ds.x.data + shift)
    lon = ds["lon"].reset_coords(drop=True)
    ds = ds.assign_coords(lon=lon + shift)
    if missing_values:
        # CESM-FV has some super high missing values. Test removal
        lon = ds["lon"].load().data
        lon[10:20, 10:20] = missing_values
        ds["lon"].data = lon

    ds_lon_corrected = correct_lon(ds)
    assert ds_lon_corrected.lon.min() >= 0
    assert ds_lon_corrected.lon.max() <= 360


def test_correct_units():
    lev = np.arange(0, 200)
    data = np.random.rand(*lev.shape)
    ds = xr.DataArray(data, dims=["lev"], coords={"lev": lev}).to_dataset(name="test")
    ds.attrs["source_id"] = "something"
    ds.lev.attrs["units"] = "centimeters"

    ds_test = correct_units(ds)
    assert ds_test.lev.attrs["units"] == "m"
    np.testing.assert_allclose(ds_test.lev.data, ds.lev.data / 100.0)


def test_correct_units_missing():
    lev = np.arange(0, 200)
    data = np.random.rand(*lev.shape)
    ds = xr.DataArray(data, dims=["lev"], coords={"lev": lev}).to_dataset(name="test")
    ds.attrs["source_id"] = "something"
    # should this raise a warning but pass?
    msg = "Unit correction failed with: Cannot convert variables"
    with pytest.warns(UserWarning, match=msg):
        ds_test = correct_units(ds)
    assert "units" not in ds_test.lev.attrs.keys()


def test_maybe_convert_bounds_to_vertex():
    # create a ds with bounds
    lon = np.arange(0, 10)
    lat = np.arange(20, 30)
    data = np.random.rand(len(lon), len(lat))
    ds = xr.DataArray(
        data, dims=["lon", "lat"], coords={"lon": lon, "lat": lat}
    ).to_dataset(name="test")
    for va in ["lon", "lat"]:
        ds.coords[va + "_bounds"] = ds[va] + xr.DataArray([-0.01, 0.01], dims=["bnds"])

    # create expected dataset
    lon_b = xr.ones_like(ds.lat) * ds.coords["lon_bounds"]
    lat_b = xr.ones_like(ds.lon) * ds.coords["lat_bounds"]

    lon_v = xr.concat(
        [lon_b.isel(bnds=ii).squeeze(drop=True) for ii in [0, 0, 1, 1]], dim="vertex"
    )
    lon_v = lon_v.reset_coords(drop=True)

    lat_v = xr.concat(
        [lat_b.isel(bnds=ii).squeeze(drop=True) for ii in [0, 1, 1, 0]], dim="vertex"
    )
    lat_v = lat_v.reset_coords(drop=True)

    ds_expected = ds.copy()
    ds_expected = ds_expected.assign_coords(lon_verticies=lon_v, lat_verticies=lat_v)

    xr.testing.assert_identical(ds_expected, maybe_convert_bounds_to_vertex(ds))
    # check that datasets that already conform to this are not changed
    xr.testing.assert_identical(
        ds_expected, maybe_convert_bounds_to_vertex(ds_expected)
    )


def test_maybe_convert_vertex_to_bounds():
    # create a ds with verticies
    lon = np.arange(0, 10)
    lat = np.arange(20, 30)
    data = np.random.rand(len(lon), len(lat))
    ds = xr.DataArray(data, dims=["x", "y"], coords={"x": lon, "y": lat}).to_dataset(
        name="test"
    )
    ds.coords["lon"] = ds.x * xr.ones_like(ds.y)
    ds.coords["lat"] = xr.ones_like(ds.x) * ds.y

    ds.coords["lon_verticies"] = (
        xr.DataArray([-0.1, -0.1, 0.1, 0.1], dims=["vertex"]) + ds["lon"]
    )
    ds.coords["lat_verticies"] = (
        xr.DataArray([-0.1, 0.1, 0.1, -0.1], dims=["vertex"]) + ds["lat"]
    )
    ds = promote_empty_dims(ds)

    # create expected dataset
    ds_expected = ds.copy()
    for va in ["lon", "lat"]:
        ds_expected.coords[va + "_bounds"] = (
            xr.DataArray([-0.1, 0.1], dims=["bnds"]) + ds_expected[va]
        )
    ds_expected = promote_empty_dims(ds_expected)

    ds_test = maybe_convert_vertex_to_bounds(ds)

    xr.testing.assert_identical(ds_expected, ds_test)
    # check that datasets that already conform to this are not changed
    xr.testing.assert_identical(
        ds_expected, maybe_convert_vertex_to_bounds(ds_expected)
    )

    assert np.all(ds_test.lon_bounds.diff("bnds") > 0)
    assert np.all(ds_test.lat_bounds.diff("bnds") > 0)


def test_sort_vertex_order():
    ordered_points = np.array([[1, 1, 2, 2], [3, 4, 4, 3]]).T

    # check every permutation of the points
    for order in list(itertools.permutations([0, 1, 2, 3])):
        points_scrambled = ordered_points[order, :]

        # create xarray
        lon_v = xr.DataArray(
            points_scrambled[:, 0],
            dims=["vertex"],
            coords={"x": 0, "y": 0},
            name="lon_bounds",
        ).expand_dims(["x", "y"])
        lat_v = xr.DataArray(
            points_scrambled[:, 1],
            dims=["vertex"],
            coords={"x": 0, "y": 0},
            name="lat_bounds",
        ).expand_dims(["x", "y"])
        da = (
            xr.DataArray([np.nan], coords={"x": 0, "y": 0})
            .expand_dims(["x", "y"])
            .to_dataset(name="test")
        )
        da = da.assign_coords({"lon_verticies": lon_v, "lat_verticies": lat_v})

        da_sorted = sort_vertex_order(da).squeeze()
        new = np.vstack((da_sorted.lon_verticies, da_sorted.lat_verticies)).T

        np.testing.assert_allclose(new, ordered_points)

        assert da_sorted.lon_verticies.isel(vertex=0) < da_sorted.lon_verticies.isel(
            vertex=3
        )
        assert da_sorted.lon_verticies.isel(vertex=1) < da_sorted.lon_verticies.isel(
            vertex=2
        )

        assert da_sorted.lat_verticies.isel(vertex=0) < da_sorted.lat_verticies.isel(
            vertex=1
        )
        assert da_sorted.lat_verticies.isel(vertex=3) < da_sorted.lat_verticies.isel(
            vertex=2
        )

        # shift the vertex by one and see if the result is the same
        da_shift = da.copy()
        da_shift = da_shift.assign_coords(vertex=da_shift.vertex + 10)
        da_sorted_shift = sort_vertex_order(da_shift).squeeze()
        np.testing.assert_allclose(da_sorted_shift.vertex.data, np.arange(4))


def test_fix_metadata():
    # Create a dataset with matching attrs
    ds = xr.Dataset()
    ds.attrs = {
        "source_id": "GFDL-CM4",
        "experiment_id": "historical",
        "branch_time_in_parent": "nonsense",
    }
    ds_fixed = fix_metadata(ds)
    assert ds_fixed.attrs["branch_time_in_parent"] == 91250

    # Test that another dataset is untouched
    ds = xr.Dataset()
    ds.attrs = {
        "source_id": "GFDL-CM4",
        "experiment_id": "other",
        "branch_time_in_parent": "nonsense",
    }
    ds_fixed = fix_metadata(ds)
    assert ds_fixed.attrs["branch_time_in_parent"] == "nonsense"


# Combination test - involving #
@pytest.mark.parametrize("add_coords", [True, False])
@pytest.mark.parametrize("shift", [0, 10])
def test_combined_preprocessing_dropped_coords(add_coords, shift):
    """Check if coordinates are properly dropped"""
    # create a 2d dataset
    xlen, ylen, zlen = (10, 5, 1)
    ds = (
        create_test_ds("x", "y", "dummy", xlen, ylen, zlen).squeeze().drop_vars("dummy")
    )
    x_bnds = xr.concat([ds.x, ds.x], "bnds")
    ds = ds.assign_coords(x_bounds=x_bnds)

    if add_coords:
        ds = ds.assign_coords(bnds=np.arange(len(ds.bnds)) + shift)

    ds = combined_preprocessing(ds)

    assert "bnds" not in ds.coords


def test_combined_preprocessing_mislabeled_coords():
    """Test if the renaming is applied to datavariables and then if they are moved to the coords."""
    # create a 2d dataset
    xlen, ylen, zlen = (10, 5, 1)
    ds = (
        create_test_ds("x", "y", "dummy", xlen, ylen, zlen).squeeze().drop_vars("dummy")
    )
    ds = ds.assign(depth=5.0)
    ds.depth.attrs["units"] = "m"  # otherwise pint complains.

    ds_pp = combined_preprocessing(ds)
    assert "lev" in ds_pp.coords
    np.testing.assert_allclose(ds.depth.data, ds_pp.lev.data)


def test_preserve_attrs():
    # create a 2d dataset
    xlen, ylen, zlen = (10, 5, 1)
    ds = (
        create_test_ds("x", "y", "dummy", xlen, ylen, zlen).squeeze().drop_vars("dummy")
    )
    ds.attrs = {"preserve_this": "here"}

    # TODO:  there are a bunch of errors if the metadata is not full.
    # I should probably ignore them and still put the datset out?
    # Well for now create one
    for att in EXACT_ATTRS:
        ds.attrs[att] = "a"

    ds_pp = combined_preprocessing(ds)
    assert ds_pp.attrs["preserve_this"] == "here"



# Preprocessing for CMIP6 models
import warnings

import cf_xarray.units  # noqa: F401
import numpy as np
import pint  # noqa: F401
import pint_xarray  # noqa: F401
import xarray as xr

from xmip.utils import _maybe_make_list, cmip6_dataset_id


# global object for units
_desired_units = {"lev": "m"}
_unit_overrides = {name: None for name in ["so"]}


_drop_coords = ["bnds", "vertex"]


def cmip6_renaming_dict():
    """a universal renaming dict. Keys correspond to source id (model name)
    and valuse are a dict of target name (key) and a list of variables that
    should be renamed into the target."""
    rename_dict = {
        # dim labels (order represents the priority when checking for the dim labels)
        "x": ["i", "ni", "xh", "nlon"],
        "y": ["j", "nj", "yh", "nlat"],
        "lev": ["deptht", "olevel", "zlev", "olev", "depth"],
        "bnds": ["bnds", "axis_nbounds", "d2"],
        "vertex": ["vertex", "nvertex", "vertices"],
        # coordinate labels
        "lon": ["longitude", "nav_lon"],
        "lat": ["latitude", "nav_lat"],
        "lev_bounds": [
            "deptht_bounds",
            "lev_bnds",
            "olevel_bounds",
            "zlev_bnds",
        ],
        "lon_bounds": [
            "bounds_lon",
            "bounds_nav_lon",
            "lon_bnds",
            "x_bnds",
            "vertices_longitude",
        ],
        "lat_bounds": [
            "bounds_lat",
            "bounds_nav_lat",
            "lat_bnds",
            "y_bnds",
            "vertices_latitude",
        ],
        "time_bounds": ["time_bnds"],
    }
    return rename_dict


def _invert_dict(rdict):
    exploded_dict = {}
    # there is probably a more effective way to 'invert' a dictionary
    for k, v in rdict.items():
        v = _maybe_make_list(v)
        for vv in v:
            exploded_dict[vv] = k
    return exploded_dict


def rename_cmip6(ds, rename_dict=None):
    """Homogenizes cmip6 dataasets to common naming"""
    ds = ds.copy()
    attrs = {k: v for k, v in ds.attrs.items()}

    if rename_dict is None:
        rename_dict = cmip6_renaming_dict()

    inverted_rename_dict = _invert_dict(rename_dict)

    ds_reset = ds.reset_coords()

    def _maybe_rename(obj, rdict):
        return obj.rename({kk: vv for kk, vv in rdict.items() if kk in obj.dims})

    # first take care of the dims and reconstruct a clean ds
    ds = xr.Dataset(
        {
            k: _maybe_rename(ds_reset[k], inverted_rename_dict)
            for k in ds_reset.data_vars
        }
    )

    # special treatment for 'lon'/'lat' if there is no 'x'/'y' after renaming process
    for di, co in [("x", "lon"), ("y", "lat")]:
        if di not in ds.dims and co in ds.dims:
            ds = ds.rename({co: di})

    # now rename the variables
    # try and pass here, cause some of the datasets (MIROC) have like 3 times the same info
    # e.g. lev/sigma/zlev...not sure this is the best way to handle this with
    # a silent fail here though...
    for va in ds.data_vars:
        try:
            ds = ds.rename({va: inverted_rename_dict[va]})
        except Exception as e:
            warnings.warn(f"Renaming failed with {e}")
            pass

    # restore attributes
    ds.attrs = attrs

    return ds


def promote_empty_dims(ds):
    """Convert empty dimensions to actual coordinates"""
    ds = ds.copy()
    for di in ds.dims:
        if di not in ds.coords:
            ds = ds.assign_coords({di: ds[di]})
    return ds


# some of the models do not have 2d lon lats, correct that.
def broadcast_lonlat(ds, verbose=True):
    """Some models (all `gr` grid_labels) have 1D lon lat arrays
    This functions broadcasts those so lon/lat are always 2d arrays."""
    if "lon" not in ds.variables:
        ds.coords["lon"] = ds["x"]
    if "lat" not in ds.variables:
        ds.coords["lat"] = ds["y"]

    if len(ds["lon"].dims) < 2:
        ds.coords["lon"] = ds["lon"] * xr.ones_like(ds["lat"])
    if len(ds["lat"].dims) < 2:
        ds.coords["lat"] = xr.ones_like(ds["lon"]) * ds["lat"]

    return ds


def _interp_nominal_lon(lon_1d):
    x = np.arange(len(lon_1d))
    idx = np.isnan(lon_1d)
    return np.interp(x, x[~idx], lon_1d[~idx], period=360)


def replace_x_y_nominal_lat_lon(ds):
    """Approximate the dimensional values of x and y with mean lat and lon at the equator"""
    ds = ds.copy()

    def maybe_fix_non_unique(data, pad=False):
        """remove duplicate values by linear interpolation
        if values are non-unique. `pad` if the last two points are the same
        pad with -90 or 90. This is only applicable to lat values"""
        if len(data) == len(np.unique(data)):
            return data
        else:
            # pad each end with the other end.
            if pad:
                if len(np.unique([data[0:2]])) < 2:
                    data[0] = -90
                if len(np.unique([data[-2:]])) < 2:
                    data[-1] = 90

            ii_range = np.arange(len(data))
            _, indicies = np.unique(data, return_index=True)
            double_idx = np.array([ii not in indicies for ii in ii_range])
            # print(f"non-unique values found at:{ii_range[double_idx]})")
            data[double_idx] = np.interp(
                ii_range[double_idx], ii_range[~double_idx], data[~double_idx]
            )
            return data

    if "x" in ds.dims and "y" in ds.dims:
        # define 'nominal' longitude/latitude values
        # latitude is defined as the max value of `lat` in the zonal direction
        # longitude is taken from the `middle` of the meridonal direction, to
        # get values close to the equator

        # pick the nominal lon/lat values from the eastern
        # and southern edge, and
        eq_idx = len(ds.y) // 2

        nominal_x = ds.isel(y=eq_idx).lon.load()
        nominal_y = ds.lat.max("x").load()

        # interpolate nans
        # Special treatment for gaps in longitude
        nominal_x = _interp_nominal_lon(nominal_x.data)
        nominal_y = nominal_y.interpolate_na("y").data

        # eliminate non unique values
        # these occour e.g. in "MPI-ESM1-2-HR"
        nominal_y = maybe_fix_non_unique(nominal_y)
        nominal_x = maybe_fix_non_unique(nominal_x)

        ds = ds.assign_coords(x=nominal_x, y=nominal_y)
        ds = ds.sortby("x")
        ds = ds.sortby("y")

        # do one more interpolation for the x values, in case the boundary values were
        # affected
        ds = ds.assign_coords(
            x=maybe_fix_non_unique(ds.x.load().data),
            y=maybe_fix_non_unique(ds.y.load().data, pad=True),
        )

    else:
        warnings.warn(
            "No x and y found in dimensions for source_id:%s. This likely means that you forgot to rename the dataset or this is the German unstructured model"
            % ds.attrs["source_id"]
        )
    return ds


def correct_units(ds):
    "Converts coordinates into SI units using pint-xarray"
    # codify units with pint
    # Perhaps this should be kept separately from the fixing?
    # See https://github.com/jbusecke/xmip/pull/160#discussion_r667041858
    try:
        # exclude salinity from the quantification (see https://github.com/jbusecke/xmip/pull/160#issuecomment-878627027 for details)
        quantified = ds.pint.quantify(_unit_overrides)
        target_units = {
            var: target_unit
            for var, target_unit in _desired_units.items()
            if var in quantified
        }

        converted = quantified.pint.to(target_units)
        ds = converted.pint.dequantify(format="~P")
    except ValueError as e:
        warnings.warn(
            f"{cmip6_dataset_id(ds)}: Unit correction failed with: {e}", UserWarning
        )
    return ds


def correct_coordinates(ds, verbose=False):
    """converts wrongly assigned data_vars to coordinates"""
    ds = ds.copy()
    for co in [
        "x",
        "y",
        "lon",
        "lat",
        "lev",
        "bnds",
        "lev_bounds",
        "lon_bounds",
        "lat_bounds",
        "time_bounds",
        "lat_verticies",
        "lon_verticies",
    ]:
        if co in ds.variables:
            if verbose:
                print("setting %s as coord" % (co))
            ds = ds.set_coords(co)
    return ds


def correct_lon(ds):
    """Wraps negative x and lon values around to have 0-360 lons.
    longitude names expected to be corrected with `rename_cmip6`"""
    ds = ds.copy()

    # remove out of bounds values found in some
    # models as missing values
    ds["lon"] = ds["lon"].where(abs(ds["lon"]) <= 1000)
    ds["lat"] = ds["lat"].where(abs(ds["lat"]) <= 1000)

    # adjust lon convention
    lon = ds["lon"].where(ds["lon"] > 0, 360 + ds["lon"])
    ds = ds.assign_coords(lon=lon)

    if "lon_bounds" in ds.variables:
        lon_b = ds["lon_bounds"].where(ds["lon_bounds"] > 0, 360 + ds["lon_bounds"])
        ds = ds.assign_coords(lon_bounds=lon_b)

    return ds


def parse_lon_lat_bounds(ds):
    """both `regular` 2d bounds and vertex bounds are parsed as `*_bounds`.
    This function renames them to `*_verticies` if the vertex dimension is found.
    Also removes time dimension from static bounds as found in e.g. `SAM0-UNICON` model.
    """
    if "source_id" in ds.attrs.keys():
        if ds.attrs["source_id"] == "FGOALS-f3-L":
            warnings.warn("`FGOALS-f3-L` does not provide lon or lat bounds.")

    ds = ds.copy()

    if "lat_bounds" in ds.variables:
        if "x" not in ds.lat_bounds.dims:
            ds.coords["lat_bounds"] = ds.coords["lat_bounds"] * xr.ones_like(ds.x)

    if "lon_bounds" in ds.variables:
        if "y" not in ds.lon_bounds.dims:
            ds.coords["lon_bounds"] = ds.coords["lon_bounds"] * xr.ones_like(ds.y)

    # I am assuming that all bound fields with time were broadcasted in error (except time bounds obviously),
    # and will drop the time dimension.
    error_dims = ["time"]
    for ed in error_dims:
        for co in ["lon_bounds", "lat_bounds", "lev_bounds"]:
            if co in ds.variables:
                if ed in ds[co].dims:
                    warnings.warn(
                        f"Found {ed} as dimension in `{co}`. Assuming this is an error and just picking the first step along that dimension."
                    )
                    stripped_coord = ds[co].isel({ed: 0}).squeeze()
                    # make sure that dimension is actually dropped
                    if ed in stripped_coord.coords:
                        stripped_coord = stripped_coord.drop(ed)

                    ds = ds.assign_coords({co: stripped_coord})

    # Finally rename the bounds that are given in vertex convention
    for va in ["lon", "lat"]:
        va_name = va + "_bounds"
        if va_name in ds.variables and "vertex" in ds[va_name].dims:
            ds = ds.rename({va_name: va + "_verticies"})

    return ds


def maybe_convert_bounds_to_vertex(ds):
    """Converts renamed lon and lat bounds into verticies, by copying
    the values into the corners. Assumes a rectangular cell."""
    ds = ds.copy()
    if "bnds" in ds.dims:
        if "lon_bounds" in ds.variables and "lat_bounds" in ds.variables:
            if (
                "lon_verticies" not in ds.variables
                and "lat_verticies" not in ds.variables
            ):
                lon_b = xr.ones_like(ds.lat) * ds.coords["lon_bounds"]
                lat_b = xr.ones_like(ds.lon) * ds.coords["lat_bounds"]

                lon_bb = xr.concat(
                    [lon_b.isel(bnds=ii).squeeze(drop=True) for ii in [0, 0, 1, 1]],
                    dim="vertex",
                )
                lon_bb = lon_bb.reset_coords(drop=True)

                lat_bb = xr.concat(
                    [lat_b.isel(bnds=ii).squeeze(drop=True) for ii in [0, 1, 1, 0]],
                    dim="vertex",
                )
                lat_bb = lat_bb.reset_coords(drop=True)

                ds = ds.assign_coords(lon_verticies=lon_bb, lat_verticies=lat_bb)

    return ds


def maybe_convert_vertex_to_bounds(ds):
    """Converts lon and lat verticies to bounds by averaging corner points
    on the appropriate cell face center."""

    ds = ds.copy()
    if "vertex" in ds.dims:
        if "lon_verticies" in ds.variables and "lat_verticies" in ds.variables:
            if "lon_bounds" not in ds.variables and "lat_bounds" not in ds.variables:
                lon_b = xr.concat(
                    [
                        ds["lon_verticies"].isel(vertex=[0, 1]).mean("vertex"),
                        ds["lon_verticies"].isel(vertex=[2, 3]).mean("vertex"),
                    ],
                    dim="bnds",
                )
                lat_b = xr.concat(
                    [
                        ds["lat_verticies"].isel(vertex=[0, 3]).mean("vertex"),
                        ds["lat_verticies"].isel(vertex=[1, 2]).mean("vertex"),
                    ],
                    dim="bnds",
                )

                ds = ds.assign_coords(lon_bounds=lon_b, lat_bounds=lat_b)
    ds = promote_empty_dims(ds)
    return ds


def sort_vertex_order(ds):
    """sorts the vertex dimension in a coherent order:
    0: lower left
    1: upper left
    2: upper right
    3: lower right
    """
    ds = ds.copy()
    if (
        "vertex" in ds.dims
        and "lon_verticies" in ds.variables
        and "lat_verticies" in ds.variables
    ):

        # pick a vertex in the middle of the domain, to avoid the pole areas
        x_idx = len(ds.x) // 2
        y_idx = len(ds.y) // 2

        lon_b = ds.lon_verticies.isel(x=x_idx, y=y_idx).load().data
        lat_b = ds.lat_verticies.isel(x=x_idx, y=y_idx).load().data
        vert = ds.vertex.load().data

        points = np.vstack((lon_b, lat_b, vert)).T

        # split into left and right
        lon_sorted = points[np.argsort(points[:, 0]), :]
        right = lon_sorted[:2, :]
        left = lon_sorted[2:, :]
        # sort again on each side to get top and bottom
        bl, tl = left[np.argsort(left[:, 1]), :]
        br, tr = right[np.argsort(right[:, 1]), :]

        points_sorted = np.vstack((bl, tl, tr, br))

        idx_sorted = (points_sorted.shape[0] - 1) - np.argsort(points_sorted[:, 2])
        ds = ds.assign_coords(vertex=idx_sorted)
        ds = ds.sortby("vertex")

    return ds


# TODO: Implement this in a sleeker way with daops
def fix_metadata(ds):
    """
    Fix known issues (from errata) with the metadata.
    """

    # https://errata.es-doc.org/static/view.html?uid=2f6b5963-f87e-b2df-a5b0-2f12b6b68d32
    if ds.attrs["source_id"] == "GFDL-CM4" and ds.attrs["experiment_id"] in [
        "1pctCO2",
        "abrupt-4xCO2",
        "historical",
    ]:
        ds.attrs["branch_time_in_parent"] = 91250
    # https://errata.es-doc.org/static/view.html?uid=61fb170e-91bb-4c64-8f1d-6f5e342ee421
    if ds.attrs["source_id"] == "GFDL-CM4" and ds.attrs["experiment_id"] in [
        "ssp245",
        "ssp585",
    ]:
        ds.attrs["branch_time_in_child"] = 60225
    return ds


def combined_preprocessing(ds):
    # fix naming
    ds = rename_cmip6(ds)
    # promote empty dims to actual coordinates
    ds = promote_empty_dims(ds)
    # demote coordinates from data_variables
    ds = correct_coordinates(ds)
    # broadcast lon/lat
    ds = broadcast_lonlat(ds)
    # shift all lons to consistent 0-360
    ds = correct_lon(ds)
    # fix the units
    ds = correct_units(ds)
    # rename the `bounds` according to their style (bound or vertex)
    ds = parse_lon_lat_bounds(ds)
    # sort verticies in a consistent manner
    ds = sort_vertex_order(ds)
    # convert vertex into bounds and vice versa, so both are available
    ds = maybe_convert_bounds_to_vertex(ds)
    ds = maybe_convert_vertex_to_bounds(ds)
    ds = fix_metadata(ds)
    ds = ds.drop_vars(_drop_coords, errors="ignore")
    return ds



import numpy as np
import xarray as xr


def _default_merge_dict():
    return {
        "North Atlantic Ocean": [
            "Caribbean Sea",
            "Gulf of Mexico",
            "Labrador Sea",
            "Hudson Bay",
            "Baffin Bay",
            "Norwegian Sea",
            "Greenland Sea",
            "Bay of Biscay",
            "Norwegian Sea",
            "Greenland Sea",
            "Gulf of Guinea",
            "Irish Sea",
            "North Sea",
            "Bahía de Campeche",
            "Davis Strait",
            "Sargasso Sea",
            "Hudson Strait",
            "English Channel",
            "Gulf of Honduras",
            "Bristol Channel",
            "Inner Seas",
            "Straits of Florida",
            "Gulf of Saint Lawrence",
            "Bay of Fundy",
            "Melville Bay",
            "Gulf of Maine",
            "Chesapeake Bay",
            "Amazon River",
            "James Bay",
            "Ungava Bay",
        ],
        "South Atlantic Ocean": ["Río de la Plata", "Golfo San Jorge"],
        "North Pacific Ocean": [
            "Philippine Sea",
            "Gulf of Alaska",
            "Sea of Okhotsk",
            "East China Sea",
            "Yellow Sea",
            "Bering Sea",
            "Golfo de California",
            "Korea Strait",
            "Cook Inlet",
            "Bristol Bay",
            "Shelikhova Gulf",
            "Bo Hai",
            "Golfo de Panamá",
            "Yangtze River",
            "Columbia River",
            "Sea of Japan",
            "Inner Sea",
        ],
        "South Pacific Ocean": [
            "Coral Sea",
            "Tasman Sea",
            "Bay of Plenty",
            "Bismarck Sea",
            "Solomon Sea",
            "Great Barrier Reef",
        ],
        "Maritime Continent": [
            "Celebes Sea",
            "Sulu Sea",
            "Banda Sea",
            "Luzon Strait",
            "Java Sea",
            "Arafura Sea",
            "Timor Sea",
            "Gulf of Thailand",
            "Gulf of Carpentaria",
            "Molucca Sea",
            "Gulf of Tonkin",
            "Strait of Malacca",
            "Strait of Singapore",
            "Makassar Strait",
            "Ceram Sea",
            "Taiwan Strait",
            "South China Sea",
        ],
        "INDIAN OCEAN": [
            "Mozambique Channel",
            "Bay of Bengal",
            "Arabian Sea",
            "Persian Gulf",
            "Andaman Sea",
            "Laccadive Sea",
            "Gulf of Aden",
            "Gulf of Oman",
            "Gulf of Mannar",
            "Gulf of Kutch",
            "Great Australian Bight",
        ],
        "Arctic Ocean": [
            "Beaufort Sea",
            "Chukchi Sea",
            "Barents Sea",
            "Kara Sea",
            "Laptev Sea",
            "White Sea",
            "The North Western Passages",
            "Amundsen Gulf",
            "Viscount Melville Sound",
        ],
        "SOUTHERN OCEAN": [
            "Ross Sea Eastern Basin",
            "Ross Sea Western Basin",
            "Weddell Sea",
            "Bellingshausen Sea",
            "Amundsen Sea",
            "Scotia Sea",
            "Drake Passage",
        ],
        "Black Sea": None,
        "Mediterranean Sea": [
            "Mediterranean Sea Eastern Basin",
            "Mediterranean Sea Western Basin",
            "Tyrrhenian Sea",
            "Adriatic Sea",
            "Golfe du Lion",
            "Ionian Sea",
            "Strait of Gibraltar",
            "Balearic Sea",
            "Aegean Sea",
        ],
        "Red Sea": None,
        "Caspian Sea": None,
        "Baltic Sea": ["Gulf of Bothnia", "Gulf of Finland"],
    }


def merged_mask(
    basins, ds, lon_name="lon", lat_name="lat", merge_dict=None, verbose=False
):
    """Combine geographical basins (from regionmask) to larger ocean basins.

    Parameters
    ----------
    basins : regionmask.core.regions.Regions object
        Loaded basin data from regionmask, e.g. `import regionmask;basins = regionmask.defined_regions.natural_earth.ocean_basins_50`
    ds : xr.Dataset
        Input dataset on which to construct the mask
    lon_name : str, optional
        Name of the longitude coordinate in `ds`, defaults to `lon`
    lat_name : str, optional
        Name of the latitude coordinate in `ds`, defaults to `lat`
    merge_dict : dict, optional
        dictionary defining new aggregated regions (as keys) and the regions to be merge into that region as as values (list of names).
        Defaults to large scale ocean basins defined by `xmip.regionmask.default_merge_dict`
    verbose : bool, optional
       Prints more output, e.g. the regions in `basins` that were not used in the merging step. Defaults to False.

    Returns
    -------
    mask : xr.DataArray
        The mask contains ascending numeric value for each key ( merged region) in `merge_dict`.
        When the default is used the numeric values correspond to the following regions:
        * 0: North Atlantic

        * 1: South Atlantic

        * 2: North Pacific

        * 3: South Pacific

        * 4: Maritime Continent

        * 5: Indian Ocean

        * 6: Arctic Ocean

        * 7: Southern Ocean

        * 8: Black Sea

        * 9: Mediterranean Sea

        *10: Red Sea

        *11: Caspian Sea

    """
    mask = basins.mask(ds, lon_name=lon_name, lat_name=lat_name)

    if merge_dict is None:
        merge_dict = _default_merge_dict()

    dict_keys = list(merge_dict.keys())
    number_dict = {k: None for k in dict_keys}
    merged_basins = []
    for ocean, small_basins in merge_dict.items():
        try:
            ocean_idx = basins.map_keys(ocean)
        except (KeyError):
            # The ocean key is new and cant be found in the previous keys (e.g. for Atlantic full or maritime continent)
            ocean_idx = mask.max().data + 1
        number_dict[ocean] = ocean_idx
        if small_basins:
            for sb in small_basins:
                sb_idx = basins.map_keys(sb)
                # set the index of each small basin to the ocean value
                mask = mask.where(mask != sb_idx, ocean_idx)
                merged_basins.append(sb)

    if verbose:
        remaining_basins = [
            str(basins.regions[ri].name)
            for ri in range(len(basins.regions))
            if (basins.regions[ri].name not in merged_basins)
            and (basins.regions[ri].name not in list(merge_dict.keys()))
        ]
        print(remaining_basins)

    # reset the mask indicies to the order of the passed dictionary keys
    mask_reordered = xr.ones_like(mask.copy()) * np.nan
    for new_idx, k in enumerate(dict_keys):
        old_idx = number_dict[k]
        mask_reordered = mask_reordered.where(mask != old_idx, new_idx)

    return mask_reordered



import warnings

import dask.array as dsa
import numpy as np
import xarray as xr
import xarrayutils as xru

from xarrayutils.utils import linear_trend

from xmip.postprocessing import EXACT_ATTRS, _match_datasets
from xmip.utils import cmip6_dataset_id


def _maybe_unpack_date(date):
    """`Unpacks` cftime from xr.Dataarray if provided"""
    # I should probably not do this here but instead in the higher level functions...
    if isinstance(date, xr.DataArray):
        date = date.data.tolist()
        if isinstance(date, list):
            if len(date) != 1:
                raise RuntimeError(
                    "The passed date has the wrong format. Got [{date}] after conversion to list."
                )
            else:
                date = date[0]
    return date


def _construct_cfdate(data, units, calendar):
    # This seems clunky. I feel there must be a more elegant way of doing this?
    date = xr.DataArray(data, attrs={"units": units, "calendar": calendar})
    return xr.decode_cf(date.to_dataset(name="time"), use_cftime=True).time


def _datestr_to_cftime(date_str, calendar):
    # Again I feel this should be more elegant? For now I guess it works
    return _construct_cfdate([0], f"days since {date_str}", calendar)


def replace_time(
    ds, ref_date=None, ref_idx=0, freq="1MS", calendar=None, time_dim="time"
):
    """This function replaces the time encoding of a dataset acoording to `ref_date`.
    The ref date can be any index of ds.time (default is 0; meaning the first timestep of ds will be replaced with `ref_date`).
    """
    # ! I might be able to achieve some of this with time.shift
    # !

    if calendar is None:
        calendar = ds.time.encoding.get("calendar", "standard")

    if ref_date is None:
        ref_date = _maybe_unpack_date(ds.time[0])

    if isinstance(ref_date, str):
        ref_date = _maybe_unpack_date(_datestr_to_cftime(ref_date, calendar))

    # TODO: Check the frequency. Currently the logic only works on monthly intervals
    if freq != "1MS":
        raise ValueError("`replace_time` currently only works with monthly data.")

    # determine the start date
    # propagate the date back (this assumes stricly monthly data)

    year = _maybe_unpack_date(ref_date).year - (ref_idx // 12)
    month = _maybe_unpack_date(ref_date).month - (ref_idx % 12)

    if month <= 0:
        # move the year one more back
        year -= 1
        month = 12 + month

    attrs = ds.time.attrs

    start = f"{int(year):04d}-{int(month):02d}"

    ds = ds.assign_coords(
        time=xr.cftime_range(start, periods=len(ds.time), freq=freq, calendar=calendar)
    )
    ds.time.attrs = attrs
    return ds


def find_date_idx(time, date):
    """Finds the index of `date` within an array of cftime dates. This strictly requires monthly data.
    Might result in undesired behavior for other time frequencies.
    """
    # ! seems like I can refactor this with http://xarray.pydata.org/en/stable/generated/xarray.CFTimeIndex.get_loc.html#xarray.CFTimeIndex.get_loc

    date = _maybe_unpack_date(date)

    # easier approach: Find the difference in years and months
    year_diff = date.year - _maybe_unpack_date(time[0]).year
    month_diff = date.month - _maybe_unpack_date(time[0]).month

    return (year_diff * 12) + month_diff


def unify_time(parent, child, adjust_to="child"):
    """Uses the CMIP6 specific metadata (augmented by xmip....time_preprocessing!!!) to adjust parent time encoding to child experiment.
    Similar to `switch_to_child_time`, but sets the time parameters (e.g. calendar) explicitly to the child conventions
    """
    branch_time_in_parent = child.attrs.get("branch_time_in_parent")

    # if branch time is not in attrs do nothing
    if branch_time_in_parent is None:
        child_source_id = child.attrs.get("source_id", "not found")
        parent_source_id = parent.attrs.get("source_id", "not found")
        msg = (
            f"Could not unify time for [child:{child_source_id}|parent:{parent_source_id}]."
            "`branch_time_in_parent` not found in attributes."
        )
        warnings.warn(msg, UserWarning)
        return parent, child

    else:
        parent_calendar = parent.time.to_index().calendar
        child_calendar = child.time.to_index().calendar
        branch_time_parent = _construct_cfdate(
            child.attrs.get("branch_time_in_parent"),
            child.attrs.get("parent_time_units"),
            parent_calendar,
        )
        branch_time_child = _construct_cfdate(
            child.attrs.get("branch_time_in_child"),
            child.time.encoding.get("units"),
            child_calendar,
        )

        if adjust_to == "child":
            branch_idx_parent = find_date_idx(parent.time, branch_time_parent)
            return (
                replace_time(
                    parent,
                    branch_time_child,
                    ref_idx=branch_idx_parent,
                    calendar=child_calendar,
                ),
                child,
            )
        elif adjust_to == "parent":
            branch_idx_child = find_date_idx(child.time, branch_time_child)
            return parent, replace_time(
                child,
                branch_time_parent,
                ref_idx=branch_idx_child,
                calendar=parent_calendar,
            )
        else:
            raise ValueError(
                f"Input for `adjust_to` not valid. Got {adjust_to}. Expected either `child` or `parent`."
            )


def calculate_drift(
    reference, ds, variable, trend_years=250, compute_short_trends=False
):
    """Calculate the linear trend at every grid position for the given time (`trend_years`)
    starting from the date when `ds` was branched of from `ds_parent`.
    CMIP6 metadata must be present.

    Parameters
    ----------
    ds_parent : xr.Dataset
        The dataset from which the drift (trend) is calculated. Usually the preindustrial control run
    ds : xr.Dataset
        The dataset for which the drift is matched. This is usually the historical experiment.
        !For many models, each historical member is branched
    trend_years : int, optional
        The duration of the trend to compute in years, by default 250 (This is the lenght of
        historical+standard scenario, e.g. 1850-2100)
    """

    for attr in [
        "parent_variant_label",
        "parent_source_id",
        "branch_time_in_parent",
        "parent_time_units",
        "source_id",
        "variant_label",
    ]:
        if attr not in ds.attrs:
            raise ValueError(f"Could not find {attr} in attributes of `ds`.")

    # Check if the parent member id matches
    match_attrs = ["source_id", "variant_label"]
    for ma in match_attrs:
        if not ds.attrs[f"parent_{ma}"] in reference.attrs[ma]:
            raise ValueError(
                f'`ds_parent` {ma} ({reference.attrs[ma]}) not compatible with `ds` parent_{ma} ({ds.attrs[f"parent_{ma}"]})'
            )

    # find the branch date in the control run
    branch_time_reference = _construct_cfdate(
        ds.attrs["branch_time_in_parent"],
        ds.attrs["parent_time_units"],
        reference.time.to_index().calendar,
    )
    branch_idx_reference = find_date_idx(reference.time, branch_time_reference)
    # there might be some cases where this is not true. Figure out what to do when it happens.
    assert branch_idx_reference >= 0

    # cut the referenmce to the appropriate time frame
    reference_cut = reference.isel(
        time=slice(branch_idx_reference, branch_idx_reference + (12 * trend_years))
    )

    if len(reference_cut.time) == 0:
        raise RuntimeError(
            "Selecting from `reference` according to the branch time resulted in empty dataset. Check the metadata."
        )
        return None
    else:
        if len(reference_cut.time) < trend_years * 12:
            if compute_short_trends:
                warnings.warn(
                    f"reference dataset does not have the full {trend_years} years to calculate trend. Using {int(len(reference_cut.time)/12)} years only"
                )
            else:
                raise RuntimeError(
                    f"Reference dataset does not have the full {trend_years} years to calculate trend. Set `calculate_short_trend=True` to compute from a shorter timeseries"
                )

        time_range = xr.concat(
            [
                reference_cut.time[0].squeeze().drop_vars("time"),
                reference_cut.time[-1].squeeze().drop_vars("time"),
            ],
            dim="bnds",
        ).reset_coords(drop=True)

        # there is some problem when encoding very large years. for now ill preserve these only as
        # strings
        time_range = time_range.astype(str)

        # The polyfit implementation actually respects the units.
        # For now my implementation requires the slope to be in units .../month
        # I might be able to change this later and accomodate other time frequencies?
        # get rid of all the additional coords, which resets the time to an integer index

        reference_cut = reference_cut[variable]

        # TODO: This has pretty poor performance...need to find out why.
        # Reset time dimension to integer index.
        #         reference_cut = reference_cut.drop_vars("time")

        # linear regression slope is all we need here.
        #         reg = reference_cut.polyfit("time", 1).sel(degree=1).polyfit_coefficients

        reg_raw = linear_trend(
            reference_cut,
            "time",
        )

        # ! quite possibly the shittiest fix ever.
        # I changed the API over at xarrayutils and now I have to pay the price over here.
        # TODO: Might want to eliminate this ones the new xarrayutils version has matured.
        if xru.__version__ > "v0.1.3":
            reg = reg_raw.slope
        else:
            reg = reg_raw.sel(parameter="slope").drop_vars("parameter").squeeze()

        # again drop all the coordinates
        reg = reg.reset_coords(drop=True)

        reg = reg.to_dataset(name=variable)

        # add metadata about regression
        reg = reg.assign_coords(trend_time_range=time_range)
        reg.coords["trend_time_range"].attrs.update(
            {
                "standard_name": "regression_time_bounds",
                "long_name": "regression_time_in_reference_run",
            }
        )
        # reg should carry the attributes of `ds`
        # ? Maybe I should convert to a dataset?
        reg.attrs.update(ds.attrs)
        return reg


# TODO: I need a more generalized detrending? Based on indicies --> xarrayutils
# Then refactor this one here just for cmip6


def detrend_basic(da, da_slope, start_idx=0, dim="time", keep_attrs=True):
    """Basic detrending just based on time index, not date"""
    # now create a trend timeseries at each point
    # and the time indicies by the ref index. This way the trend is correctly calculated from the reference year.
    # this adapts the chunk structure from the input if its a dask array
    attrs = {k: v for k, v in da.attrs.items()}
    idx_start = -start_idx
    idx_stop = len(da.time) - start_idx
    if isinstance(da.data, dsa.Array):
        ref_time = da.isel({di: 0 for di in da.dims if di != dim})
        chunks = ref_time.chunks
        trend_time_idx_data = dsa.arange(
            idx_start, idx_stop, chunks=chunks, dtype=da.dtype
        )
    else:
        trend_time_idx_data = np.arange(idx_start, idx_stop, dtype=da.dtype)

    trend_time_idx = xr.DataArray(
        trend_time_idx_data,
        dims=[dim],
    )

    # chunk like the time dimension
    slope = da_slope.squeeze()

    trend = trend_time_idx * slope

    detrended = da - trend
    if keep_attrs:
        detrended.attrs.update(attrs)
    return detrended


def remove_trend(ds, ds_slope, variable, ref_date, check_mask=True):
    """Detrending method for cmip6 data. Only works with monthly data!
    This does not correct the time convention. Be careful with experiements that have
    a non compatible time convention (often control runs.)
    """

    if not isinstance(ds, xr.Dataset):
        raise ValueError("`ds` input needs to be a dataset")

    if not isinstance(ds_slope, xr.Dataset):
        raise ValueError("`ds_slope` input needs to be a dataset")

    da = ds[variable]
    da_slope = ds_slope[variable]

    da, da_slope = xr.align(da, da_slope, join="override")

    if check_mask:
        nanmask_data = np.isnan(da.isel(time=[0, len(da.time) // 2, -1])).all("time")
        nanmask_slope = np.isnan(da_slope)
        # perform a quick test to see if the land is aligned properly
        if np.logical_xor(nanmask_data, nanmask_slope).any():
            raise ValueError(
                "Nanmask between data and slope array not identical. Check input and disable `check_mask` to skip this test"
            )

    ref_calendar = da.time.to_index().calendar
    ref_date = xr.cftime_range(ref_date, periods=1, calendar=ref_calendar)

    # Find the index corresponding to the ref date (this can be outside the range of the actual data)
    ref_idx = find_date_idx(da.time, ref_date)

    detrended = detrend_basic(
        da, da_slope, start_idx=ref_idx, dim="time", keep_attrs=True
    )

    # add information to track which data was used to remove trend
    if "trend_time_range" in ds_slope.coords:
        trend_start = ds_slope.trend_time_range.isel(bnds=0).load().data.tolist()
        trend_stop = ds_slope.trend_time_range.isel(bnds=1).load().data.tolist()

    else:
        trend_start = "not-available"
        trend_stop = "not-available"
        warnings.warn(
            "`ds_slope` did not have information about the time over which the slope was calculated. Check the input."
        )

    detrended.attrs[
        "drift_removed"
    ] = f"linear_trend_{cmip6_dataset_id(ds_slope)}_{trend_start}_{trend_stop}"

    return detrended


def match_and_remove_trend(
    ddict, trend_ddict, ref_date="1850", nomatch="warn", **detrend_kwargs
):
    """Find and remove trend files from a dictonary of datasets

    Parameters
    ----------
    ddict : dict
        dictionary with xr.Datasets which should get a trend/drift removed
    trend_ddict : dict
        dictionary with results of linear regressions. These should be removed from the datasets in `ddict`
    ref_date : str, optional
        Start date of the trend, by default "1850"
    nomatch : str, optional
        Define the behavior when for a given dataset in `ddict` there is no matching trend dataset in `trend_ddict`.
        Can be `warn`, `raise`, or `ignore`, by default 'warn'

    Returns
    -------
    dict
        Dictionary of detrended dataasets. Only contains values of `ddict` that actually had a trend removed.

    """
    ddict_detrended = {}
    match_attrs = [ma for ma in EXACT_ATTRS if ma not in ["experiment_id"]] + [
        "variable_id"
    ]

    for k, ds in ddict.items():
        trend_ds = _match_datasets(
            ds, trend_ddict, match_attrs, pop=False, unique=True, nomatch=nomatch
        )
        if len(trend_ds) == 2:
            trend_ds = trend_ds[
                1
            ]  # this is a bit clunky. _match_datasest does return the input ds, so we have to grab the second one?
            # I guess I could pass *trend_ds, but that is not very readable
            variable = ds.attrs["variable_id"]
            da_detrended = ds.assign(
                {
                    variable: remove_trend(
                        ds, trend_ds, variable, ref_date=ref_date, **detrend_kwargs
                    )
                }
            )
            # should this just return a dataset instead?
            ddict_detrended[k] = da_detrended

    return ddict_detrended



import functools
import inspect
import warnings

from typing import List, Mapping

import numpy as np
import xarray as xr

from xmip.utils import _key_from_attrs, _maybe_make_list, cmip6_dataset_id


try:
    import xesmf
except ImportError:
    xesmf = None


# define the attrs that are needed to get an 'exact' match
EXACT_ATTRS = [
    "source_id",
    "grid_label",
    "experiment_id",
    "table_id",
    # "version", # for testing
    "variant_label",
    "variable_id",
]


def _match_attrs(ds_a, ds_b, match_attrs):
    """returns the number of matched attrs between two datasets"""
    try:
        n_match = sum([ds_a.attrs[i] == ds_b.attrs[i] for i in match_attrs])
        return n_match
    except KeyError:
        raise ValueError(
            f"Cannot match datasets because at least one of the datasets does not contain all attributes [{match_attrs}]."
        )


def _match_datasets(ds, ds_dict, match_attrs, pop=True, nomatch="ignore", unique=False):
    """Find all datasets in a dictionary of datasets that have a matching set
    of attributes and return a list of datasets for merging/concatting.
    Optionally remove the matching datasets from the input dict.
    """
    datasets = [ds]
    keys = list(ds_dict.keys())
    for k in keys:
        if _match_attrs(ds, ds_dict[k], match_attrs) == len(match_attrs):
            if pop:
                ds_matched = ds_dict.pop(k)
            else:
                ds_matched = ds_dict[k]
            # preserve the original dictionary key of the chosen dataset in attribute.
            ds_matched.attrs["original_key"] = k
            datasets.append(ds_matched)
    if len(datasets) > 2:
        if unique:
            raise ValueError(
                f"Found more than one matching dataset for {cmip6_dataset_id(ds)}. Pass `unique=False` to ignore this."
            )
    nomatch_msg = f"Could not find a matching dataset for {cmip6_dataset_id(ds)}"
    if len(datasets) < 2:
        if nomatch == "ignore":
            pass
        elif nomatch == "warn":
            warnings.warn(nomatch_msg)
        elif nomatch == "raise":
            raise RuntimeError(nomatch_msg)
        else:
            # Could this be done in an annotation? Or some other built in way to do this?
            raise ValueError(
                f"Invalid input ({nomatch}) for `nomatch`, should be `ignore`, `warn`, or `raise`"
            )
    return datasets


def _prune_match_attrs_to_available(
    match_attrs: List[str], ds_dict: Mapping[str, xr.Dataset]
) -> List[str]:
    """prune a set of attrs to only the ones available in every dataset"""
    missing_match_attrs = []
    for ma in match_attrs:
        if all([ma not in ds.attrs.keys() for ds in ds_dict.values()]):
            missing_match_attrs.append(ma)

    if len(missing_match_attrs) > 0:
        warnings.warn(
            f"Match attributes {missing_match_attrs} not found in any of the datasets. \
        This can happen when several combination functions are used and attributes are removed during merging. \
        Double check the results."
        )
        pruned_match_attrs = [ma for ma in match_attrs if ma not in missing_match_attrs]
        return pruned_match_attrs
    else:
        return match_attrs


def combine_datasets(
    ds_dict,
    combine_func,
    combine_func_args=(),
    combine_func_kwargs={},
    match_attrs=EXACT_ATTRS,
):
    """General combination function to combine datasets within a dictionary according to their matching attributes.
    This function provides maximum flexibility, but can be somewhat complex to set up. The postprocessing module provided several
    convenience wrappers like `merge_variables`, `concat_members`, etc.

    Parameters
    ----------
    ds_dict : [type]
        [description]
    combine_func : [type]
        [description]
    combine_func_args : tuple, optional
        [description], by default ()
    combine_func_kwargs : dict, optional
        [description], by default {}
    match_attrs : [type], optional
        [description], by default exact_attrs

    Returns
    -------
    [type]
        [description]
    """
    # make a copy of the input dict, so it is not modified outside of the function
    # ? Not sure this is always desired.
    ds_dict = {k: v for k, v in ds_dict.items()}
    ds_dict_combined = {}

    # Check each of the matching attributes. If attr is not present in any of
    # the input datasets drop from match_attrs and warn
    match_attrs = _prune_match_attrs_to_available(match_attrs, ds_dict)

    while len(ds_dict) > 0:
        # The order does not matter here, so just choose the first key
        k = list(ds_dict.keys())[0]
        ds = ds_dict.pop(k)
        matched_datasets = _match_datasets(ds, ds_dict, match_attrs, pop=True)

        # create new dict key
        new_k = _key_from_attrs(ds, match_attrs, sep=".")

        # for now Ill hardcode the merging. I have been thinking if I should
        # just pass a general `func` to deal with a list of datasets. The user could pass custom stuff
        # And I think that way I can generalize all/most of the `member` level combine functions.
        # Well for now, lets do it the manual way.
        try:
            ds_combined = combine_func(
                matched_datasets, *combine_func_args, **combine_func_kwargs
            )
            ds_dict_combined[new_k] = ds_combined
        except Exception as e:
            warnings.warn(f"{cmip6_dataset_id(ds)} failed to combine with :{e}")

    return ds_dict_combined


# Convenience Wrappers for `combine_datasets`. Less flexible but easier to use and understand.
def merge_variables(
    ds_dict,
    merge_kwargs={},
):
    """Given a dictionary of datasets, this function merges all available data variables (given in seperate datasets) into a single dataset.
    CAUTION: This assumes that all variables are on the same staggered grid position. If you are working with data on the cell edges,
    this function will disregard that information. Use the grids module instead to get an accurate staggered grid representation.

    Parameters
    ----------
    ds_dict : dict
        Dictionary of xarray datasets, that need matching metrics.
    merge_kwargs : dict
        Optional arguments passed to xr.merge.

    Returns
    -------
    dict
        A new dict of xr.Datasets with all datasets from `ds_dict`, but with merged variables and adjusted keys.

    """
    # An exact match is only possible if all attrs in `EXACT_ATTRS` are in every dataset

    match_attrs = [ma for ma in EXACT_ATTRS if ma not in ["variable_id"]]

    # set defaults
    merge_kwargs.setdefault("compat", "override")
    merge_kwargs.setdefault("join", "exact")  # if the size differs throw an error.
    merge_kwargs.setdefault(
        "combine_attrs", "drop_conflicts"
    )  # if the size differs throw an error. Requires xarray >=0.17.0

    return combine_datasets(
        ds_dict, xr.merge, combine_func_kwargs=merge_kwargs, match_attrs=match_attrs
    )


def _construct_and_promote_member_id(ds):
    # construct member_id according to https://docs.google.com/document/d/1h0r8RZr_f3-8egBMMh7aqLwy3snpD6_MrDz1q8n5XUk/edit
    sub_experiment_id = ds.attrs.get("sub_experiment_id", "none")
    variant_label = ds.attrs[
        "variant_label"
    ]  # if this should fail in the future, we could build an error/warning in here
    if sub_experiment_id == "none":
        member_id = variant_label
    else:
        member_id = f"{sub_experiment_id}-{variant_label}"

    if "member_id" not in ds.dims:
        ds = ds.expand_dims({"member_id": [member_id]})
    else:
        existing_member_id = ds.member_id.data[0]
        if not existing_member_id == member_id:
            warnings.warn(
                f"{cmip6_dataset_id(ds)} already contained a member_id ({existing_member_id}) but this is different from the reconstructed value ({member_id}). The existing value is not modified, but we recommend checking the input.",
                UserWarning,
            )
    return ds


def concat_members(
    ds_dict,
    concat_kwargs={},
):
    """Given a dictionary of datasets, this function merges all available ensemble members
    (given in seperate datasets) into a single dataset for each combination of attributes,
    like source_id, grid_label, etc. but with concatnated members.
    CAUTION: If members do not have the same dimensions (e.g. longer run time for some members),
    this can result in poor dask performance (see: https://github.com/jbusecke/xmip/issues/58)

    Parameters
    ----------
    ds_dict : dict
        Dictionary of xarray datasets.
    concat_kwargs : dict
        Optional arguments passed to xr.concat.

    Returns
    -------
    dict
        A new dict of xr.Datasets with all datasets from `ds_dict`, but with concatenated members and adjusted keys.

    """
    match_attrs = [ma for ma in EXACT_ATTRS if ma not in ["variant_label"]]

    # set defaults
    concat_kwargs.setdefault(
        "combine_attrs", "drop_conflicts"
    )  # if the size differs throw an error. Requires xarray >=0.17.0

    # promote variant_label attr to coordinate, to have member_ids as coordinates

    ds_dict = {k: _construct_and_promote_member_id(ds) for k, ds in ds_dict.items()}

    return combine_datasets(
        ds_dict,
        xr.concat,
        combine_func_args=(
            ["member_id"]
        ),  # I dont like this. Its confusing to have two different dimension names
        combine_func_kwargs=concat_kwargs,
        match_attrs=match_attrs,
    )


def _concat_sorted_time(ds_list, **kwargs):
    # extract the first date
    start_dates = [str(ds.time.to_index()[0]) for ds in ds_list]
    sorted_idx = np.argsort(start_dates)
    ds_list_sorted = [ds_list[i] for i in sorted_idx]
    return xr.concat(ds_list_sorted, "time", **kwargs)


def concat_experiments(
    ds_dict,
    concat_kwargs={},
):
    """Given a dictionary of datasets, this function merges all available ensemble members
    (given in seperate datasets) into a single dataset for each combination of attributes,
    like source_id, grid_label, etc. but with concatnated members.
    CAUTION: If members do not have the same dimensions (e.g. longer run time for some members),
    this can result in poor dask performance (see: https://github.com/jbusecke/xmip/issues/58)
    Parameters
    ----------
    ds_dict : dict
        Dictionary of xarray datasets.
    concat_kwargs : dict
        Optional arguments passed to xr.concat.
    Returns
    -------
    dict
        A new dict of xr.Datasets with all datasets from `ds_dict`, but with concatenated members and adjusted keys.
    """

    match_attrs = [ma for ma in EXACT_ATTRS if ma not in ["experiment_id"]]

    # set defaults
    concat_kwargs.setdefault(
        "combine_attrs",
        "drop_conflicts",
    )  # if the size differs throw an error. Requires xarray >=0.17.0
    concat_kwargs.setdefault("compat", "override")
    concat_kwargs.setdefault("coords", "minimal")

    return combine_datasets(
        ds_dict,
        _concat_sorted_time,
        combine_func_kwargs=concat_kwargs,
        match_attrs=match_attrs,
    )


def _pick_first_member(ds_list, **kwargs):
    members = [ds.variant_label for ds in ds_list]
    first_member_idx = np.argmin(members)
    return ds_list[first_member_idx]


def pick_first_member(ddict):
    match_attrs = [ma for ma in EXACT_ATTRS if ma not in ["variant_label"]]
    return combine_datasets(
        ddict,
        _pick_first_member,
        match_attrs=match_attrs,
    )


# Matching wrapper specific to combining grid labels via interpolation with xesmf
def requires_xesmf(func):
    @functools.wraps(func)
    def wrapper_requires_xesmf(*args, **kwargs):

        # Check if xesmf is installed
        if xesmf is None:
            raise ValueError(
                "This function needs the optional package xesmf. Please install with `conda install -c conda-forge xesmf`."
            )
        return func(*args, **kwargs)

    return wrapper_requires_xesmf


def _pick_grid_label(ds_list, preferred_grid_label):
    """From a list of datasets this picks the dataset with `ds.attrs['grid_label']==preferred_label`
    if available"""
    # TODO: Implement an option to pick the highest/lowest resolution if grid label dows not match.
    matches = [ds for ds in ds_list if ds.attrs["grid_label"] == preferred_grid_label]
    if len(matches) > 0:
        return matches[0]
    else:
        return ds_list[0]


def _drop_duplicate_grid_labels(ddict, preferred_grid_label):
    """Eliminate multiple grid labels for the same dataset, by dropping multiples
    with a preference for `preferred_grid_label`"""
    match_attrs = [ma for ma in EXACT_ATTRS if ma not in ["version", "grid_label"]] + [
        "variable_id"
    ]
    return combine_datasets(
        ddict,
        _pick_grid_label,
        combine_func_args=(preferred_grid_label,),
        match_attrs=match_attrs,
    )


def _clean_regridder(ds_source, ds_target, method, **xesmf_kwargs):
    def _clean(ds):
        # remove all unnecessary stuff for the regridding
        ds = ds.isel(time=0, lev=0, rho=0, missing_dims="ignore")
        for coord in [
            co for co in ds.coords if co not in ["lon", "lat"]
        ]:  # , "lat_bounds""lon_bounds",
            ds = ds.drop_vars(coord)

        # Ugly Hack to elminate 'seam' when regridding from gr to native grids
        # There is something in xesmf that causes problems with broadcasted regular lon/lat values
        if "gr" in ds.attrs["grid_label"]:
            # actually revert the convention and return the 1d coordinates. ?Should I change that behavior in general?
            ds = ds.assign(lon=ds.lon.isel(y=0))
            ds = ds.assign(lat=ds.lat.isel(x=0))

        # for now just eliminate the attrs here
        # I can solve this more elegantly when I parse proper cf-attributes
        for coord in ds.coords:
            ds[coord].attrs = {}
        #         ds = ds.rename({'lon_bounds':'lon_b', 'lat_bounds':'lat_b'})

        # TODO: Make this work out of the box with lon/lat bounds and method='conservative'
        # TODO: Maybe erase the need for this completely with cf-xarray
        return ds

    ds_source = _clean(ds_source)
    ds_target = _clean(ds_target)
    return xesmf.Regridder(ds_source, ds_target, method, **xesmf_kwargs)


def _regrid_to_target(ds_source, ds_target, regridder):
    ds_regridded = regridder(ds_source, keep_attrs=True)

    # remove all coordinates that involve x and y (these will be merged from the native grid dataset)
    ds_regridded = ds_regridded.reset_coords(drop=True)

    # modify attributes
    ds_regridded.attrs["grid_label"] = ds_target.attrs["grid_label"]
    # identify the variables that are regridded
    for var in ds_regridded.data_vars:
        ds_regridded[var].attrs["xmip_regrid_method"] = ds_regridded.attrs[
            "regrid_method"
        ]

    return ds_regridded


def _interpolate_combine_func(
    ds_list, target_grid_label, method, verbose=False, xesmf_kwargs={}, merge_kwargs={}
):
    grid_labels_raw = [dss.attrs["grid_label"] for dss in ds_list]
    grid_labels = np.unique(grid_labels_raw)

    # special case: All grid labels are the same but none of them is equal to target_grid_label
    if len(grid_labels) == 1:
        return xr.merge(ds_list, **merge_kwargs)
    else:

        if target_grid_label not in grid_labels:
            raise ValueError(
                f"Could not find any variable with the target_grid_label ({target_grid_label}). Found these instead: {grid_labels}"
            )
        else:
            # just take the first one with a matching grid?
            target_grid = [
                ds for ds in ds_list if ds.attrs["grid_label"] == target_grid_label
            ][0]

            # Construct a regridder for each other grid_label
            regridder_dict = {}
            if verbose:
                print(
                    f'Constructing regridders for source_id {target_grid.attrs["source_id"]} ...'
                )
            for gl in grid_labels:
                if gl != target_grid_label:
                    if verbose:
                        print(gl)
                    source_grid = [
                        dss for dss in ds_list if dss.attrs["grid_label"] == gl
                    ][
                        0
                    ]  # again just take the first one available
                    regridder_dict[gl] = _clean_regridder(
                        source_grid, target_grid, method, **xesmf_kwargs
                    )
            if verbose:
                print("FINISHED")

            # Now regrid all datasets in the list (dont do anything if already on the target_grid)
            ds_list_new = []
            for ds_raw in ds_list:
                if ds_raw.attrs["grid_label"] != target_grid_label:
                    if verbose:
                        print(f"regridding {cmip6_dataset_id(ds_raw)}")
                    ds_regridded = _regrid_to_target(
                        ds_raw, target_grid, regridder_dict[ds_raw.attrs["grid_label"]]
                    )
                    ds_list_new.append(ds_regridded)
                else:
                    ds_list_new.append(ds_raw)

            # check that horizontal dimensions are compatible.
            xy_dimensions = [(len(ds.x), len(ds.y)) for ds in ds_list_new]
            if not all([xy == xy_dimensions[0] for xy in xy_dimensions]):
                raise ValueError(
                    f"Regridded datasets do not have the same dimensions. Found({xy_dimensions}). This will cause broadcasting problems during merge."
                )
            if verbose:
                print("Merging regridded files")
            return xr.merge(ds_list_new, **merge_kwargs)


@requires_xesmf
def interpolate_grid_label(
    ds_dict,
    target_grid_label="gn",
    method="bilinear",
    xesmf_kwargs={},
    merge_kwargs={},
    verbose=False,
):
    """Combines different grid labels via interpolation with xesmf

    Parameters
    ----------
    ds_dict : dict
        dictonary of input datasets
    target_grid_label : str, optional
        preferred grid_label value. If at least one dataset has this grid_label, otherse are interpolated to it.
        Dataset with this grid label are not modified, by default "gn"
    method : str, optional
        interpolation method for xesmf, by default "bilinear"
    xesmf_kwargs : dict, optional
        optional arguments for building xesmf regridder, by default {}
    merge_kwargs : dict, optional
        optional arguments for the merging of interpolated datasets, by default {}
    verbose : bool, optional
        print output while creating regridder, by default False

    Returns
    -------
    dict
        dictionary of combined datasets (usually will combine across different variable ids)
    """
    match_attrs = [
        ma for ma in EXACT_ATTRS if ma not in ["grid_label", "variable_id"]
    ]  # does this need to be more flexible?

    xesmf_kwargs.setdefault("ignore_degenerate", True)
    xesmf_kwargs.setdefault("periodic", True)

    merge_kwargs.setdefault("combine_attrs", "drop_conflicts")

    # first drop the datasets that might have both the target and another grid_label present
    ds_dict = _drop_duplicate_grid_labels(ds_dict, target_grid_label)

    return combine_datasets(
        ds_dict,
        _interpolate_combine_func,
        combine_func_args=(
            target_grid_label,
            method,
        ),
        combine_func_kwargs={
            "xesmf_kwargs": xesmf_kwargs,
            "merge_kwargs": merge_kwargs,
        },
        match_attrs=match_attrs,
    )


# Matching wrapper specific to metric datasets


def _parse_metric(ds, metric, dim_length_conflict="error"):
    """Convenience function to parse a metric dataarry into an existing dataset.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset
    metric : xr.Data
        Metric dataarray to add to `ds`
    dim_length_conflict : str
        Specifies how to handle dimensions that are not the same lenght.

    Returns
    -------
    xr.Dataset
        `ds` with `metric` added as coordinate.

    """
    dataset_id = cmip6_dataset_id(ds)

    if not isinstance(metric, xr.DataArray):
        raise ValueError(
            f"{dataset_id}:`metric` input must be xarray.DataArray. Got {type(metric)}"
        )

    if metric.name is None:
        warnings.warn(
            f"{dataset_id}:`metric` has no name. This might lead to problems down the line.",
            RuntimeWarning,
        )

    # Check dimension compatibility
    # TODO: Check if xarray has some nifty way to do this
    mismatch_dims = []
    for di in metric.dims:
        if len(ds[di]) != len(metric[di]):
            mismatch_dims.append(di)

    if len(mismatch_dims) > 0:
        metric_str = [str(di) + ":" + str(len(metric[di])) for di in mismatch_dims]
        ds_str = [str(di) + ":" + str(len(ds[di])) for di in mismatch_dims]
        msg = (
            f"{dataset_id}:`metric` dimensions {metric_str} do not match `ds` {ds_str}."
        )
        if dim_length_conflict == "error":
            raise ValueError(msg)
        elif dim_length_conflict == "align":
            warnings.warn(msg + " Aligning the data on `inner`")
            ds, metric = xr.align(ds, metric, join="inner")

    # strip all coordinates from metric
    metric_stripped = metric.reset_coords(drop=True)

    # add attributes
    metric_stripped.attrs[
        "parsed_with"
    ] = f"xmip/postprocessing/{inspect.currentframe().f_code.co_name}"
    # TODO: Get the package and module name without hardcoding.

    ds = ds.assign_coords({metric_stripped.name: metric_stripped})

    return ds


def match_metrics(
    ds_dict,
    metric_dict,
    match_variables,
    match_attrs=["source_id", "grid_label"],
    print_statistics=False,
    dim_length_conflict="error",
):
    """Given two dictionaries of datasets, this function matches metrics from `metric_dict` to
    every datasets in `ds_dict` based on a comparison of the datasets attributes.

    Parameters
    ----------
    ds_dict : dict
        Dictionary of xarray datasets, that need matching metrics.
    metric_dict : dict
        Dictionary of xarray datasets, which contain metrics as data_variables.
    match_variables : list
        Data variables of datasets in `metric_dict` to parse.
    match_attrs : list, optional
        Minimum dataset attributes that need to match, by default ["source_id", "grid_label"].
        Pass "exact" to only allow exact matches using all required attributes.
    print_statistics : bool, optional
        Option to print statistics about matching, by default False
    dim_length_conflict : str
        Defines the behavior when parsing metrics with non-exact matches in dimension size.
        See `parse_metric`.
    Returns
    -------
    dict
        All datasets from `ds_dict`, if match was not possible the input dataset is returned unchanged.

    """
    # metrics should never match the variable
    exact_attrs_wo_var = [ma for ma in EXACT_ATTRS if ma != "variable_id"]

    # TODO: this naming is a big weird. Basically this here is necessary to still get a 'closest match' on whichever of the full
    # set of 'exact' attrs are given on each of the datasets
    pruned_attrs = _prune_match_attrs_to_available(exact_attrs_wo_var, ds_dict)
    # and this also needs to be true for the metrics of course.
    pruned_attrs = _prune_match_attrs_to_available(pruned_attrs, metric_dict)

    match_variables = _maybe_make_list(match_variables)

    # if match is set to exact check all these attributes
    if match_attrs == "exact":
        match_attrs = exact_attrs_wo_var

    # --- set up statistics --- #
    total_datasets = 0
    exact_datasets = {va: 0 for va in match_variables}
    nonexact_datasets = {va: 0 for va in match_variables}
    nomatch_datasets = {va: 0 for va in match_variables}

    ds_dict_parsed = {}
    for k, ds in ds_dict.items():
        total_datasets += 1
        # Filter for the matching attrs
        matched_metrics = _match_datasets(ds, metric_dict, match_attrs, pop=False)

        # Filter for the variable
        for mv in match_variables:
            matched_metric_vars = [ds for ds in matched_metrics if mv in ds.variables]

            if len(matched_metric_vars) == 0:
                warnings.warn(f"No matching metrics found for {mv}")
                nomatch_datasets[mv] += 1
            else:
                # Pick the metric with the most mathching attributes.
                # This ensures that the exact match is picked if available.

                nmatch = [
                    _match_attrs(ds, ds_match, pruned_attrs)
                    for ds_match in matched_metric_vars
                ]
                closest_match = np.argmax(nmatch)
                ds_metric = matched_metric_vars[closest_match]

                # this is a hardcoded check for time variable metrics.
                # These are very likely only valid for exact matches of runs.
                # For instance these could be cell thickness values, which cannot
                # simply be used for all runs of a model.
                exact_match = _match_attrs(ds, ds_metric, pruned_attrs) == len(
                    exact_attrs_wo_var
                )
                if "time" in ds_metric.dims and not exact_match:
                    warnings.warn(
                        "This metric had a time dimension and did not perfectly match. Not parsing anything."
                    )
                else:
                    # parse the original key attr to the dataarray
                    ds_metric[mv].attrs["original_key"] = ds_metric.attrs[
                        "original_key"
                    ]
                    ds = _parse_metric(
                        ds, ds_metric[mv], dim_length_conflict=dim_length_conflict
                    )
                    if exact_match:
                        exact_datasets[mv] += 1
                    else:
                        nonexact_datasets[mv] += 1

        ds_dict_parsed[k] = ds
    if print_statistics:
        print(
            f"Processed {total_datasets} datasets.\nExact matches:{exact_datasets}\nOther matches:{nonexact_datasets}\nNo match found:{nomatch_datasets}"
        )
    return ds_dict_parsed



import inspect

import cftime
import numpy as np
import pytest
import xarray as xr

from xmip.drift_removal import (
    _construct_cfdate,
    calculate_drift,
    find_date_idx,
    match_and_remove_trend,
    remove_trend,
    replace_time,
    unify_time,
)
from xmip.postprocessing import EXACT_ATTRS
from xmip.utils import cmip6_dataset_id


# I copied this from a PR I made to parcels a while back.
# Is there a more elegant way to parse these from cftime?
def _get_cftime_datetimes():
    cftime_calendars = tuple(
        x[1].__name__ for x in inspect.getmembers(cftime._cftime, inspect.isclass)
    )
    cftime_datetime_names = [ca for ca in cftime_calendars if "Datetime" in ca]
    return cftime_datetime_names


all_cftime_calendars = [
    getattr(cftime, cf_datetime)(1990, 1, 1).calendar
    for cf_datetime in _get_cftime_datetimes()
]


@pytest.mark.parametrize("source_calendar", all_cftime_calendars)
@pytest.mark.parametrize("target_calendar", all_cftime_calendars)
@pytest.mark.parametrize(
    "ref_idx, start_date",
    [
        (-12, "2001-01-01"),
        (12, "1999-01-01"),
        (1, "1999-12-01"),
        (11, "1999-02-01"),
        (13, "1998-12-01"),
        (-13, "2001-02-01"),
    ],
)
def test_replace_time(source_calendar, target_calendar, ref_idx, start_date):
    nt = 10
    timename = "time"
    time = xr.cftime_range(
        "2000-01-01", periods=nt, freq="1MS", calendar=source_calendar
    )
    ds = xr.DataArray(
        np.random.rand(nt), dims=timename, coords={timename: time}
    ).to_dataset(name="test")

    # simply replace with a different calendar
    time_replaced = xr.DataArray(
        xr.cftime_range("2000-01-01", periods=nt, freq="1MS", calendar=target_calendar),
        dims=["time"],
    )
    ds_replaced = replace_time(ds, calendar=target_calendar)
    xr.testing.assert_allclose(time_replaced, ds_replaced.time)

    # shift time with reference date before the actual data (e.g. for forced runs)
    time_shifted = xr.DataArray(
        xr.cftime_range(start_date, periods=nt, freq="1MS", calendar=target_calendar),
        dims=["time"],
    )
    ds_shifted = replace_time(
        ds, calendar=target_calendar, ref_idx=ref_idx, ref_date="2000-01-01"
    )
    xr.testing.assert_allclose(time_shifted, ds_shifted.time)


def test_replace_time_error_other_freq():
    # Check that a useful error is raised if the input frequency is not monthly
    nt = 10
    timename = "time"
    time = xr.cftime_range("2000-01-01", periods=nt, freq="1MS")
    ds = xr.DataArray(
        np.random.rand(nt), dims=timename, coords={timename: time}
    ).to_dataset(name="test")
    with pytest.raises(
        ValueError, match="replace_time` currently only works with monthly data."
    ):
        replace_time(ds, freq="1DS")


@pytest.mark.parametrize(
    "expected_idx, strdate",
    [(0, "2000-01-01"), (0, "2000-01-29"), (2, "2000-03-29"), (-12, "1999-01-01")],
)
@pytest.mark.parametrize("calendar", all_cftime_calendars)
def test_find_date_idx(expected_idx, strdate, calendar):
    time = xr.cftime_range("2000-01-01", periods=10, freq="1MS")
    date = _construct_cfdate([0], f"days since {strdate}", calendar)
    assert find_date_idx(time, date) == expected_idx


@pytest.mark.parametrize("source_calendar", all_cftime_calendars)
@pytest.mark.parametrize("target_calendar", all_cftime_calendars)
def test_unify_time_simple(source_calendar, target_calendar):
    nt = 24
    # simplest case (same length, just different time convention)
    time_child = xr.DataArray(
        xr.cftime_range("2000-01-01", periods=nt, freq="1MS", calendar=source_calendar),
        dims=["time"],
    )
    time_parent = xr.DataArray(
        xr.cftime_range("1900-01-01", periods=nt, freq="1MS", calendar=target_calendar),
        dims=["time"],
    )

    ds_child = xr.DataArray(
        np.random.rand(12, nt),
        dims=["x", "time"],
        coords={"time": time_child},
    ).to_dataset(name="test")

    ds_child.attrs = {
        "branch_time_in_parent": 0,
        "branch_time_in_child": 0,
        "parent_time_units": "days since 1900-01-01",
    }
    # this should be written when files are actually read from a file, but this is
    # a bit dangerous. Why is the child time units not set?
    ds_child.time.encoding["units"] = "days since 2000-01-01"

    ds_parent = xr.DataArray(
        np.random.rand(12, nt), dims=["x", "time"], coords={"time": time_parent}
    ).to_dataset(name="test")

    ds_parent_adjusted, ds_child_adjusted = unify_time(ds_parent, ds_child)

    xr.testing.assert_allclose(ds_parent_adjusted.time, time_child)

    ds_parent_adjusted, ds_child_adjusted = unify_time(
        ds_parent, ds_child, adjust_to="parent"
    )
    xr.testing.assert_allclose(ds_child_adjusted.time, time_parent)


@pytest.mark.parametrize("source_calendar", all_cftime_calendars)
@pytest.mark.parametrize("target_calendar", all_cftime_calendars)
def test_unify_time_complex(source_calendar, target_calendar):
    nt = 24
    nt_parent = 200
    # more complex and common case (longer control, with starting from 0 convention)
    time_child = xr.DataArray(
        xr.cftime_range("2000-01-01", periods=nt, freq="1MS", calendar=source_calendar),
        dims=["time"],
    )
    time_parent = xr.DataArray(
        xr.cftime_range(
            "0100-01-01", periods=nt_parent, freq="1MS", calendar=target_calendar
        ),
        dims=["time"],
    )

    ds_child = xr.DataArray(
        np.random.rand(12, nt),
        dims=["x", "time"],
        coords={"time": time_child},
    ).to_dataset(name="test")

    ds_child.attrs = {
        "branch_time_in_parent": 0,
        "branch_time_in_child": 0,
        "parent_time_units": "days since 0200-01-01",
    }
    # this should be written when files are actually read from a file, but this is
    # a bit dangerous. Why is the child time units not set?
    ds_child.time.encoding["units"] = "days since 2000-01-01"

    ds_parent = xr.DataArray(
        np.random.rand(12, nt_parent), dims=["x", "time"], coords={"time": time_parent}
    ).to_dataset(name="test")

    ds_parent_adjusted, ds_child_adjusted = unify_time(ds_parent, ds_child)

    expected_time_parent_adjusted = xr.DataArray(
        xr.cftime_range(
            "1900-01-01", periods=nt_parent, freq="1MS", calendar=source_calendar
        ),
        dims=["time"],
    )

    xr.testing.assert_allclose(ds_parent_adjusted.time, expected_time_parent_adjusted)

    ds_parent_adjusted, ds_child_adjusted = unify_time(
        ds_parent, ds_child, adjust_to="parent"
    )

    expected_time_child_adjusted = xr.DataArray(
        xr.cftime_range("0200-01-01", periods=nt, freq="1MS", calendar=target_calendar),
        dims=["time"],
    )
    xr.testing.assert_allclose(ds_child_adjusted.time, expected_time_child_adjusted)


def test_unify_time_missing_attr_warning():
    nt = 24
    # simplest case (same length, just different time convention)
    time_child = xr.cftime_range("2000-01-01", periods=nt, freq="1MS")
    time_parent = xr.cftime_range("1200-01-01", periods=nt, freq="1MS")

    ds_child = xr.DataArray(
        np.random.rand(12, nt), dims=["x", "time"], coords={"time": time_child}
    )
    ds_parent = xr.DataArray(
        np.random.rand(12, nt), dims=["x", "time"], coords={"time": time_parent}
    )
    with pytest.warns(UserWarning):
        ds_parent_adjusted, ds_child_adjusted = unify_time(ds_parent, ds_child)
    xr.testing.assert_equal(ds_parent, ds_parent_adjusted)
    xr.testing.assert_equal(ds_child, ds_child_adjusted)


def test_unify_time_adjust_to_error():
    nt = 24
    # simplest case (same length, just different time convention)
    time_child = xr.DataArray(
        xr.cftime_range("2000-01-01", periods=nt, freq="1MS"),
        dims=["time"],
    )
    time_parent = xr.DataArray(
        xr.cftime_range("1900-01-01", periods=nt, freq="1MS"),
        dims=["time"],
    )

    ds_child = xr.DataArray(
        np.random.rand(12, nt),
        dims=["x", "time"],
        coords={"time": time_child},
    ).to_dataset(name="test")

    ds_child.attrs = {
        "branch_time_in_parent": 0,
        "branch_time_in_child": 0,
        "parent_time_units": "days since 1900-01-01",
    }
    # this should be written when files are actually read from a file, but this is
    # a bit dangerous. Why is the child time units not set?
    ds_child.time.encoding["units"] = "days since 2000-01-01"

    ds_parent = xr.DataArray(
        np.random.rand(12, nt), dims=["x", "time"], coords={"time": time_parent}
    ).to_dataset(name="test")
    with pytest.raises(ValueError):
        ds_parent_adjusted, ds_child_adjusted = unify_time(
            ds_parent, ds_child, adjust_to="nonsense"
        )


@pytest.mark.parametrize("chunk", [False, {"x": -1, "bnds": 1}])
def test_remove_trend(chunk):

    # normal testing
    time = xr.cftime_range("1850-01-01", periods=5, freq="1AS")
    data = xr.DataArray(
        np.random.rand(3, 4, len(time)),
        dims=["x", "y", "time"],
        coords={"time": time},
        attrs={"just_some": "test"},
    )

    slope = xr.DataArray(np.random.rand(3, 4), dims=["x", "y"])

    ref_date = str(time[0])
    dummy_time = xr.DataArray(np.arange(len(time)), dims=["time"])

    sloped_data = data + (slope * dummy_time)
    sloped_data.attrs = data.attrs

    sloped_data = sloped_data.to_dataset(name="test")
    slope = slope.to_dataset(name="test")
    time_range = xr.DataArray(
        [
            "test_start",
            "test_stop",
        ],
        dims="bnds",
    )
    slope = slope.assign_coords(trend_time_range=time_range)
    if chunk:
        slope = slope.chunk(chunk)

    detrended = remove_trend(sloped_data, slope, "test", ref_date)
    xr.testing.assert_allclose(data, detrended)
    for att in data.attrs.keys():
        assert detrended.attrs[att] == data.attrs[att]

    assert (
        detrended.attrs["drift_removed"]
        == f"linear_trend_{cmip6_dataset_id(slope)}_test_start_test_stop"
    )

    # test the additional output when the slope input does not have sufficient information

    with pytest.warns(UserWarning):
        detrended = remove_trend(
            sloped_data, slope.drop_vars("trend_time_range"), "test", ref_date
        )
    assert (
        detrended.attrs["drift_removed"]
        == f"linear_trend_{cmip6_dataset_id(slope)}_not-available_not-available"
    )


@pytest.mark.parametrize("chunk", [False, {"time": 2}])
def test_remove_trend_mask_check(chunk):

    time = xr.cftime_range("1850-01-01", periods=5, freq="1AS")
    data = xr.DataArray(
        np.random.rand(3, 4, len(time)), dims=["x", "y", "time"], coords={"time": time}
    )

    slope = xr.DataArray(
        np.random.rand(3, 4), dims=["x", "y"], attrs={"reference_year": str(time[0])}
    )

    slope[0, 0] = np.nan

    ref_date = str(time[0])
    with pytest.raises(ValueError):
        remove_trend(data, slope, "test", ref_date)


@pytest.mark.parametrize("chunk", [False, {"time": 2}])
def test_remove_trend_exceptions(chunk):

    # normal testing
    time = xr.cftime_range("1850-01-01", periods=5, freq="1AS")
    data = xr.DataArray(
        np.random.rand(3, 4, len(time)),
        dims=["x", "y", "time"],
        coords={"time": time},
        attrs={"just_some": "test"},
    )
    slope = xr.DataArray(np.random.rand(3, 4), dims=["x", "y"])

    ref_date = str(time[0])
    with pytest.raises(ValueError) as einfo:
        remove_trend(data, slope.to_dataset(name="test"), "test", ref_date)
    assert str(einfo.value) == "`ds` input needs to be a dataset"

    with pytest.raises(ValueError) as einfo:
        remove_trend(data.to_dataset(name="test"), slope, "test", ref_date)
    assert str(einfo.value) == "`ds_slope` input needs to be a dataset"


def test_calculate_drift_missing_attrs():
    # error if no attr are given
    ds_control = xr.DataArray([0]).to_dataset(name="test")
    ds = xr.DataArray([0]).to_dataset(name="test")
    with pytest.raises(ValueError) as einfo:
        calculate_drift(ds_control, ds, "test", trend_years=250)
    assert "in attributes of `ds`." in str(einfo.value)

    # error for attrs mismatch
    ds.attrs = {
        "source_id": "a",
        "variant_label": "b",
        "branch_time_in_parent": 0,
        "parent_time_units": "something",
        "parent_variant_label": "a",
        "parent_source_id": "a",
    }

    ds_control.attrs = {
        "source_id": "a",
        "variant_label": "b",
    }

    with pytest.raises(ValueError) as einfo:
        calculate_drift(ds_control, ds, "test", trend_years=250)
    assert (
        str(einfo.value)
        == "`ds_parent` variant_label (b) not compatible with `ds` parent_variant_label (a)"
    )


@pytest.mark.parametrize("trend_years", [1, 5, 10])
def test_calculate_drift(trend_years):
    # error if no attr are given
    nx, ny = (10, 20)
    nt_control = 3000
    nt = 24
    time_control = xr.cftime_range("0100-01-01", periods=nt_control, freq="1MS")
    time_ds = xr.cftime_range("2000-01-01", periods=nt, freq="1MS")

    ds_control = xr.DataArray(
        np.random.rand(nx, ny, nt_control),
        dims=["x", "y", "time"],
        coords={"time": time_control},
    ).to_dataset(name="test")

    ds = xr.DataArray(
        np.random.rand(nx, ny, nt),
        dims=["x", "y", "time"],
        coords={"time": time_ds},
    ).to_dataset(name="test")

    ds.attrs = {
        "source_id": "a",
        "variant_label": "a",
        "branch_time_in_parent": 0,
        "parent_time_units": "days since 0105-01-01",
        "parent_variant_label": "a",
        "parent_source_id": "a",
    }

    ds_control.attrs = {
        "source_id": "a",
        "variant_label": "a",
    }

    reg = calculate_drift(ds_control, ds, "test", trend_years=trend_years)

    # use times from the output to cut control and manually calculate expected values
    start = reg.trend_time_range.isel(bnds=0).data.tolist()
    stop = reg.trend_time_range.isel(bnds=1).data.tolist()

    # Need to replace the time with an index to ensure the units of the drift are in ../month.
    ds_control_expected_normed = ds_control.sel(time=slice(start, stop))
    ds_control_expected_normed = ds_control_expected_normed.assign_coords(
        time=np.arange(len(ds_control_expected_normed.time))
    )

    reg_expected = (
        (
            ds_control_expected_normed.test.polyfit("time", 1)
            .sel(degree=1)
            .polyfit_coefficients
        )
        .drop_vars(["x", "y", "degree"])
        .squeeze()
    )

    xr.testing.assert_allclose(reg_expected, reg.test)
    assert reg.attrs == ds.attrs

    # TODO: Assert the correct time limits in the coords


def test_calculate_drift_exceptions():
    # error if no attr are given
    nx, ny = (10, 20)
    nt_control = 30
    nt = 24
    time_control = xr.cftime_range("0100-01-01", periods=nt_control, freq="1MS")
    time_ds = xr.cftime_range("2000-01-01", periods=nt, freq="1MS")

    ds_control = xr.DataArray(
        np.random.rand(nx, ny, nt_control),
        dims=["x", "y", "time"],
        coords={"time": time_control},
    ).to_dataset(name="test")

    ds = xr.DataArray(
        np.random.rand(nx, ny, nt),
        dims=["x", "y", "time"],
        coords={"time": time_ds},
    ).to_dataset(name="test")

    ds.attrs = {
        "source_id": "a",
        "variant_label": "a",
        "branch_time_in_parent": 0,
        "parent_time_units": "days since 0605-01-01",
        "parent_variant_label": "a",
        "parent_source_id": "a",
    }

    ds_control.attrs = {
        "source_id": "a",
        "variant_label": "a",
    }
    msg = "Selecting from `reference` according to the branch time resulted in empty dataset. Check the metadata."
    with pytest.raises(RuntimeError, match=msg):
        calculate_drift(ds_control, ds, "test")


def test_calculate_drift_exceptions_partial():
    # error if no attr are given
    nx, ny = (10, 20)
    nt_control = 24
    nt = 24
    time_control = xr.cftime_range("0100-01-01", periods=nt_control, freq="1MS")
    time_ds = xr.cftime_range("2000-01-01", periods=nt, freq="1MS")

    ds_control = xr.DataArray(
        np.random.rand(nx, ny, nt_control),
        dims=["x", "y", "time"],
        coords={"time": time_control},
    ).to_dataset(name="test")

    ds = xr.DataArray(
        np.random.rand(nx, ny, nt),
        dims=["x", "y", "time"],
        coords={"time": time_ds},
    ).to_dataset(name="test")

    ds.attrs = {
        "source_id": "a",
        "variant_label": "a",
        "branch_time_in_parent": 0,
        "parent_time_units": "days since 0101-01-01",
        "parent_variant_label": "a",
        "parent_source_id": "a",
    }

    ds_control.attrs = {
        "source_id": "a",
        "variant_label": "a",
    }
    msg = "Set `calculate_short_trend=True` to compute from a shorter timeseries"
    with pytest.raises(RuntimeError, match=msg):
        calculate_drift(ds_control, ds, "test")
    # TODO: Assert the correct time limits in the attrs

    with pytest.warns(UserWarning) as winfo:
        calculate_drift(ds_control, ds, "test", compute_short_trends=True)
    assert any(
        [
            "years to calculate trend. Using 1 years only" in w.message.args[0]
            for w in winfo
        ]
    )


@pytest.mark.parametrize("ref_date", ["1850", "2000-01-02"])
def test_match_and_remove_trend_matching_experiment(ref_date):
    # construct a dict of data to be detrended
    nx, ny = (10, 20)
    nt = 24
    time_historical = xr.cftime_range("1850-01-01", periods=nt, freq="1MS")
    time_ssp = xr.cftime_range("2014-01-01", periods=nt, freq="1MS")
    raw_attrs = {k: "dummy" for k in EXACT_ATTRS + ["variable_id"]}

    ds_a_hist_vara = xr.DataArray(
        np.random.rand(nx, ny, nt),
        dims=["x", "y", "time"],
        coords={"time": time_historical},
    ).to_dataset(name="vara")
    ds_a_hist_vara.attrs = {k: v for k, v in raw_attrs.items()}
    ds_a_hist_vara.attrs["variant_label"] = "a"
    ds_a_hist_vara.attrs["variable_id"] = "vara"
    ds_a_hist_vara.attrs["experiment_id"] = "historical"

    ds_a_hist_varb = xr.DataArray(
        np.random.rand(nx, ny, nt),
        dims=["x", "y", "time"],
        coords={"time": time_historical},
    ).to_dataset(name="varb")
    ds_a_hist_varb.attrs = {k: v for k, v in raw_attrs.items()}
    ds_a_hist_varb.attrs["variant_label"] = "a"
    ds_a_hist_varb.attrs["variable_id"] = "varb"
    ds_a_hist_varb.attrs["experiment_id"] = "historical"
    ds_a_other_vara = xr.DataArray(
        np.random.rand(nx, ny, nt),
        dims=["x", "y", "time"],
        coords={"time": time_ssp},
    ).to_dataset(name="vara")
    ds_a_other_vara.attrs = {k: v for k, v in raw_attrs.items()}
    ds_a_other_vara.attrs["variant_label"] = "a"
    ds_a_other_vara.attrs["variable_id"] = "vara"
    ds_a_other_vara.attrs["experiment_id"] = "other"

    ds_b_hist_vara = xr.DataArray(
        np.random.rand(nx, ny, nt),
        dims=["x", "y", "time"],
        coords={"time": time_historical},
    ).to_dataset(name="vara")
    ds_b_hist_vara.attrs = {k: v for k, v in raw_attrs.items()}
    ds_b_hist_vara.attrs["variant_label"] = "b"
    ds_b_hist_vara.attrs["variable_id"] = "vara"
    ds_b_hist_vara.attrs["experiment_id"] = "historical"

    ddict = {
        "ds_a_hist_vara": ds_a_hist_vara,
        "ds_a_hist_varb": ds_a_hist_varb,
        "ds_a_other_vara": ds_a_other_vara,
        "ds_b_hist_vara": ds_b_hist_vara,
    }

    trend_a_vara = (
        xr.ones_like(ds_a_hist_vara.isel(time=0)).drop_vars("time") * np.random.rand()
    )
    trend_a_varb = (
        xr.ones_like(ds_a_hist_varb.isel(time=0)).drop_vars("time") * np.random.rand()
    )

    trend_b_vara = (
        xr.ones_like(ds_b_hist_vara.isel(time=0)).drop_vars("time") * np.random.rand()
    )
    # trend_b_varb = (
    #     xr.ones_like(ds_b_hist_varb.isel(time=0)).drop_vars("time") * np.random.rand()
    # )
    # print(trend_b_varb)

    ddict_trend = {
        n: ds for n, ds in enumerate([trend_a_vara, trend_b_vara, trend_a_varb])
    }

    ddict_detrended = match_and_remove_trend(ddict, ddict_trend, ref_date=ref_date)

    for name, ds, trend_ds in [
        ("ds_a_hist_vara", ds_a_hist_vara, trend_a_vara),
        ("ds_b_hist_vara", ds_b_hist_vara, trend_b_vara),
        ("ds_a_hist_varb", ds_a_hist_varb, trend_a_varb),
        ("ds_a_other_vara", ds_a_other_vara, trend_a_vara),
    ]:
        variable = ds.attrs["variable_id"]
        expected = remove_trend(ds, trend_ds, variable, ref_date).to_dataset(
            name=variable
        )
        xr.testing.assert_allclose(
            ddict_detrended[name],
            expected,
        )


def test_match_and_remove_trend_nomatch():
    # create two datasets that do not match (according to the hardcoded conventions in `match_and_detrend`)
    ds = xr.DataArray().to_dataset(name="test")
    ds.attrs = {k: "a" for k in EXACT_ATTRS + ["variable_id"]}
    ds_nomatch = xr.DataArray().to_dataset(name="test")
    ds_nomatch.attrs = {k: "b" for k in EXACT_ATTRS + ["variable_id"]}

    detrended = match_and_remove_trend({"aa": ds}, {"bb": ds_nomatch}, nomatch="ignore")
    assert detrended == {}

    match_msg = "Could not find a matching dataset for *"
    with pytest.warns(UserWarning, match=match_msg):
        detrended = match_and_remove_trend(
            {"aa": ds}, {"bb": ds_nomatch}, nomatch="warn"
        )

    with pytest.raises(RuntimeError, match=match_msg):
        detrended = match_and_remove_trend(
            {"aa": ds}, {"bb": ds_nomatch}, nomatch="raise"
        )


def test_match_and_remove_trend_nonunique():
    # create two datasets that do not match (according to the hardcoded conventions in `match_and_detrend`)
    ds = xr.DataArray().to_dataset(name="test")
    ds.attrs = {k: "a" for k in EXACT_ATTRS + ["variable_id"]}
    ds_match_a = xr.DataArray().to_dataset(name="test")
    ds_match_b = xr.DataArray().to_dataset(name="test")
    ds_match_a.attrs = ds.attrs
    ds_match_b.attrs = ds.attrs

    match_msg = "Found more than one matching dataset for *"
    with pytest.raises(ValueError, match=match_msg):
        match_and_remove_trend({"aa": ds}, {"bb": ds_match_a, "cc": ds_match_b})
